"""Subtract input B from input A statement."""

from __future__ import (unicode_literals, print_function, absolute_import)
from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from iotile.sg.exceptions import SensorGraphSemanticError
from ...stream import DataStream


@python_2_unicode_compatible
class SubtractStatement(SensorGraphStatement):
    """Subtract the two input values

    The form of the statement should be
    subtract stream => stream[, default value];

    Where the default value is optional and is used to initialize the
    constant stream.  If it is not passed it defaults to 0.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        location (LocationInfo): A namedtuple with information on the line this
            statement was generated from so that we can log appropriate error
            messages.
    """

    def __init__(self, parsed, location=None):
        self.subtract_stream = parsed[0]
        self.stream = parsed[1]
        self.default = None

        if 'default' in parsed:
            self.default = parsed['default']

        super(SubtractStatement, self).__init__([], location)

    def __str__(self):
        default = u""
        if self.default is not None:
            default = ", default %d" % self.default

        return 'subtract {} => {}{};'.format(self.subtract_stream, self.stream, default)

    def execute(self, sensor_graph, scope_stack):
        """Execute this statement on the sensor_graph given the current scope tree.

        This adds a single node to the sensor graph with subtract as the function
        so that the current scope's trigger stream has the subtract_stream's value
        subtracted from it.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        if self.subtract_stream.stream_type != DataStream.ConstantType:
            raise SensorGraphSemanticError("You can only subtract a constant value currently", stream=self.subtract_stream)

        parent = scope_stack[-1]
        alloc = parent.allocator

        trigger_stream, trigger_cond = parent.trigger_chain()

        sensor_graph.add_node(u"({} always && {} {}) => {} using {}".format(self.subtract_stream, trigger_stream, trigger_cond, self.stream, 'subtract_a_from_b'))

        value = self.default
        if value is None:
            value = 0

        if self.default is not None and self.subtract_stream in sensor_graph.constant_database:
            raise SensorGraphSemanticError("Attempted to set the same constant stream twice", stream=self.subtract_stream, new_value=self.default)
        elif self.default is None and self.subtract_stream in sensor_graph.constant_database:
            return

        sensor_graph.add_constant(self.subtract_stream, value)



"""A meta variable definition that is used for passing information along with a sensor graph."""

from future.utils import python_2_unicode_compatible
from iotile.sg.exceptions import SensorGraphSemanticError
from .statement import SensorGraphStatement
from ..scopes import RootScope


@python_2_unicode_compatible
class MetaStatement(SensorGraphStatement):
    """A metadata variable definition.

    The statement form is:
    meta var = value;

    where var must be an identifier and value must be an rvalue.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        location (LocationInfo): A namedtuple with information on the line this
            statement was generated from so that we can log appropriate error
            messages.
    """

    def __init__(self, parsed, location=None):
        self.identifier = parsed[0]
        self.value = parsed[1]

        super(MetaStatement, self).__init__([], location)

    def __str__(self):
        return u'meta %s = %s;' % (str(self.identifier), str(self.value))

    def execute(self, sensor_graph, scope_stack):
        """Execute this statement on the sensor_graph given the current scope tree.

        This function will likely modify the sensor_graph and will possibly
        also add to or remove from the scope_tree.  If there are children nodes
        they will be called after execute_before and before execute_after,
        allowing block statements to sandwich their children in setup and teardown
        functions.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        if not isinstance(scope_stack[-1], RootScope):
            raise SensorGraphSemanticError("You may only declare metadata at global scope in a sensorgraph.", identifier=self.identifier, value=self.value)

        sensor_graph.add_metadata(self.identifier, self.value)



from .scope import Scope
from ...node import InputTrigger
from ...exceptions import SensorGraphSemanticError
from ...stream import DataStream

class TriggerScope(Scope):
    """A scope that implements trigger_chain but cannot provide a clock.

    Args:
        sensor_graph (SensorGraph): The sensor graph we are working on.
        scope_stack (list(Scope)): The stack of already allocated scopes.
        trigger_input ((DataStream, InputTrigger): The input trigger stream
            and condition.  The input trigger must already be attached.
    """

    def __init__(self, sensor_graph, scope_stack, trigger_input):
        parent = scope_stack[-1]
        alloc = parent.allocator
        sensor_graph = parent.sensor_graph

        super(TriggerScope, self).__init__(u"Trigger Scope", sensor_graph, alloc, parent)

        # Create our own node to create our triggering chain
        # this will be optimized out if it turned out not to be needed
        # after all nodes have been allocated.  Since we can trigger on a root
        # node, make sure we don't try to make our output a root input.
        stream_type = trigger_input[0].stream_type
        if stream_type == DataStream.InputType:
            stream_type = DataStream.UnbufferedType

        stream = alloc.allocate_stream(stream_type)
        sensor_graph.add_node(u'({} {}) => {} using copy_latest_a'.format(trigger_input[0], trigger_input[1], stream))
        self.trigger_stream = stream
        self.trigger_cond = InputTrigger(u'count', '==', 1)

    def trigger_chain(self):
        """Return a NodeInput tuple for creating a node.

        Returns:
            (StreamIdentifier, InputTrigger)
        """

        trigger_stream = self.allocator.attach_stream(self.trigger_stream)
        return (trigger_stream, self.trigger_cond)

    def clock(self, interval, basis):
        """Return a NodeInput tuple for triggering an event every interval.

        We request each distinct type of clock at most once and combine it with our
        latch stream each time it is requested.

        Args:
            interval (int): The interval (in seconds) at which this input should
                trigger.
        """

        raise SensorGraphSemanticError("TriggerScope does not provide a clock output that can be used.")



"""Stream walkers are the basic data retrieval mechanism in sensor graph."""

from iotile.core.exceptions import ArgumentError, InternalError
from iotile.sg.exceptions import StreamEmptyError
from iotile.sg.stream import DataStream


class StreamWalker(object):
    """An object that walks through the SensorLog and returns matching readings.

    Args:
        selector (DataStreamSelector): The selector for the stream(s) that we
            are walking.

    Returns:
        bool: Whether the walker matches the stream
    """

    def __init__(self, selector):
        self.selector = selector

    def matches(self, stream):
        """Check if a stream matches this walker.

        Args:
            stream (DataStream): The stream to check
        """

        return self.selector.matches(stream)

    @property
    def buffered(self):
        return self.selector.buffered


class BufferedStreamWalker(StreamWalker):
    """A stream walker backed by a storage buffer.

    Args:
        selector (DataStreamSelector): The selector for the streams
            that we are walking
        engine (StorageEngine): The storage engine backing us up
    """

    def __init__(self, selector, engine):
        super(BufferedStreamWalker, self).__init__(selector)
        self.engine = engine
        self._count = 0

        storage, streaming = self.engine.count()

        if selector.output:
            self.offset = streaming
            self.storage_type = u'streaming'
        else:
            self.offset = storage
            self.storage_type = u'storage'

    def count(self):
        return self._count

    def pop(self):
        """Pop a reading off of this stream and return it."""

        if self._count == 0:
            raise StreamEmptyError("Pop called on buffered stream walker without any data", selector=self.selector)

        while True:
            curr = self.engine.get(self.storage_type, self.offset)
            self.offset += 1

            stream = DataStream.FromEncoded(curr.stream)
            if self.matches(stream):
                self._count -= 1
                return curr

    def peek(self):
        """Peek at the oldest reading in this virtual stream."""

        if self._count == 0:
            raise StreamEmptyError("Peek called on buffered stream walker without any data", selector=self.selector)

        offset = self.offset

        while True:
            curr = self.engine.get(self.storage_type, offset)
            offset += 1

            stream = DataStream.FromEncoded(curr.stream)
            if self.matches(stream):
                return curr

    def skip_all(self):
        """Skip all readings in this walker."""

        storage, streaming = self.engine.count()

        if self.selector.output:
            self.offset = streaming
        else:
            self.offset = storage

        self._count = 0

    def notify_added(self, stream):
        """Notify that a new reading has been added.

        Args:
            stream (DataStream): The stream that had new data
        """

        if not self.matches(stream):
            return

        self._count += 1

    def notify_rollover(self, stream):
        """Notify that a reading in the given stream was overwritten.

        Args:
            stream (DataStream): The stream that had overwritten data.
        """

        self.offset -= 1

        if not self.matches(stream):
            return

        if self._count == 0:
            raise InternalError("BufferedStreamWalker out of sync with storage engine, count was wrong.")

        self._count -= 1


class VirtualStreamWalker(StreamWalker):
    """A stream walker that just walks over virtual streams.

    Args:
        selector (DataStreamSelector): The selector for the stream(s) that we
            are walking.
    """

    def __init__(self, selector):
        super(VirtualStreamWalker, self).__init__(selector)

        if selector.match_id is None:
            raise ArgumentError("You cannot create a stream walker with a wildcard virtual stream")

        self.reading = None

    def push(self, stream, value):
        """Update this stream walker with a new responsive reading.

        Virtual stream walkers keep at most one reading so this function
        just overwrites whatever was previously stored.
        """

        if not self.matches(stream):
            raise ArgumentError("Attempting to push reading to stream walker that does not match", selector=self.selector, stream=stream)

        self.reading = value

    def iter(self):
        """Iterate over the readings that are responsive to this stream walker."""

        if self.reading is not None:
            yield self.reading

    def count(self):
        if self.selector.match_type == DataStream.ConstantType:
            return 0xFFFFFFFF

        if self.reading is None:
            return 0

        return 1

    def pop(self):
        """Pop a reading off of this virtual stream and return it."""

        if self.reading is None:
            raise StreamEmptyError("Pop called on virtual stream walker without any data", selector=self.selector)

        reading = self.reading

        # If we're not a constant stream, we just exhausted ourselves
        if self.selector.match_type != DataStream.ConstantType:
            self.reading = None

        return reading

    def peek(self):
        """Peek at the oldest reading in this virtual stream."""

        if self.reading is None:
            raise StreamEmptyError("Pop called on virtual stream walker without any data", selector=self.selector)

        return self.reading

    def skip_all(self):
        """Skip all readings in this walker."""

        # We can't skip a constant stream
        if self.selector.match_type == DataStream.ConstantType:
            return

        self.reading = None


class CounterStreamWalker(StreamWalker):
    """A stream walker that walks across a counter stream.

    Counter streams only store one reading but keep an accurate
    count of how many times they've been pushed and popped and
    return the last value pushed for each each pop until they're
    empty.

    Args:
        selector (DataStreamSelector): The selector for the stream(s) that
            we're supposed to walk.
    """

    def __init__(self, selector):
        super(CounterStreamWalker, self).__init__(selector)

        if selector.match_id is None:
            raise ArgumentError("You cannot create a stream walker with a wildcard virtual stream")

        self.reading = None
        self._count = 0

    def push(self, stream, value):
        """Update this stream walker with a new responsive reading.

        Args:
            stream (DataStream): The stream that we're pushing
            value (IOTileReading): The reading tha we're pushing
        """

        if not self.matches(stream):
            raise ArgumentError("Attempting to push reading to stream walker that does not match", selector=self.selector, stream=stream)

        self.reading = value
        self._count += 1

    def iter(self):
        """Iterate over the readings that are responsive to this stream walker."""

        for i in xrange(0, self._count):
            yield self.reading

    def count(self):
        """Count how many readings are available in this walker."""

        return self._count

    def pop(self):
        """Pop a reading off of this virtual stream and return it."""

        if self._count == 0:
            raise StreamEmptyError("Pop called on virtual stream walker without any data", selector=self.selector)

        self._count = self._count - 1
        return self.reading

    def peek(self):
        """Peek at the oldest reading in this virtual stream."""

        if self.reading is None:
            raise StreamEmptyError("peek called on virtual stream walker without any data", selector=self.selector)

        return self.reading

    def skip_all(self):
        """Skip all readings in this walker."""

        self._count = 0


class InvalidStreamWalker(StreamWalker):
    """A stream walker that is not connected to anything.

    These stream walkers are cannot be used to hold any data and always
    have a count() of 0 but they are useful for unconnected inputs on
    sensor graph node.  The only functions that work are:

        skip_all
        count

    Args:
        selector (DataStreamSelector): The selector for the stream(s) that
            we're supposed to walk.
    """

    def __init__(self, selector):
        super(InvalidStreamWalker, self).__init__(selector)

    def matches(self, stream):
        """Check if a stream matches this walker.

        Args:
            stream (DataStream): The stream to check

        Returns:
            bool: Whether the walker matches the stream
        """

        return False

    def push(self, stream, value):
        """Update this stream walker with a new responsive reading.

        Args:
            stream (DataStream): The stream that we're pushing
            value (IOTileReading): The reading tha we're pushing
        """

        raise ArgumentError("Attempting to push reading to an invalid stream walker that cannot hold data", selector=self.selector, stream=stream)


    def iter(self):
        """Iterate over the readings that are responsive to this stream walker."""

        return []

    def count(self):
        """Count how many readings are available in this walker."""

        return 0

    def pop(self):
        """Pop a reading off of this virtual stream and return it."""

        raise StreamEmptyError("Pop called on an invalid stream walker")

    def peek(self):
        """Peek at the oldest reading in this virtual stream."""

        raise StreamEmptyError("peek called on an invalid stream walker")

    def skip_all(self):
        """Skip all readings in this walker."""

        pass



"""Output formats that we can save a sensor graph in."""

from .snippet import format_snippet
from .ascii import format_ascii
from .config import format_config

known_formats = {
    'snippet': format_snippet,
    'ascii': format_ascii,
    'config': format_config
}


__all__ = [u'known_formats']



"""Base class for RPC executors.

RPC Executors are objects whose job is to send an RPC and interpret
the response as a sensor graph reading.

They can either run the RPC on an actual device or they can run it
locally in a simulated device or just mock the response as needed.
"""

import struct
from iotile.core.exceptions import InternalError, HardwareError


class RPCExecutor(object):
    """RPC Executors run RPCs on behalf of sensor graph."""

    def __init__(self):
        self.warning_channel = None
        self.mock_rpcs = {}

    def mock(self, slot, rpc_id, value):
        """Store a mock return value for an RPC

        Args:
            slot (SlotIdentifier): The slot we are mocking
            rpc_id (int): The rpc we are mocking
            value (int): The value that should be returned
                when the RPC is called.
        """

        address = slot.address

        if address not in self.mock_rpcs:
            self.mock_rpcs[address] = {}

        self.mock_rpcs[address][rpc_id] = value

    def warn(self, message):
        """Let the user of this RPCExecutor know about an issue.

        warn should be used when the RPC processed in a nonstandard way
        that typically indicates a misconfiguration or problem but did
        technically complete successfully.
        """

        if self.warning_channel is None:
            return

    def rpc(self, address, rpc_id):
        """Call an RPC and receive the result as an integer.

        If the RPC does not properly return a 32 bit integer, raise a warning
        unless it cannot be converted into an integer at all, in which case
        a HardwareError is thrown.

        Args:
            address (int): The address of the tile we want to call the RPC
                on
            rpc_id (int): The id of the RPC that we want to call

        Returns:
            int: The result of the RPC call.  If the rpc did not succeed
                an error is thrown instead.
        """

        # Always allow mocking an RPC to override whatever the defaul behavior is
        if address in self.mock_rpcs and rpc_id in self.mock_rpcs[address]:
            value = self.mock_rpcs[address][rpc_id]
            return value

        result = self._call_rpc(address, rpc_id, bytes())

        if len(result) != 4:
            self.warn(u"RPC 0x%X on address %d: response had invalid length %d not equal to 4" % (rpc_id, address, len(result)))

        if len(result) < 4:
            raise HardwareError("Response from RPC was not long enough to parse as an integer", rpc_id=rpc_id, address=address, response_length=len(result))

        if len(result) > 4:
            result = result[:4]

        res, = struct.unpack("<L", result)
        return res

    def _call_rpc(self, address, rpc_id, payload):
        """Call an RPC with the given information and return its response.

        Must raise a hardware error of the appropriate kind if the RPC
        can not be executed correctly.  Otherwise it should return the binary
        response payload received from the RPC.

        Args:
            address (int): The address of the tile we want to call the RPC
                on
            rpc_id (int): The id of the RPC that we want to call
            payload (bytes, bytearray): The data that we want to send as the payload
        """

        raise InternalError("RPCExecutor did not properly override _call_rpc")



"""Remove nodes whose output is not used."""

from __future__ import unicode_literals, absolute_import, print_function
from iotile.core.exceptions import ArgumentError
from iotile.sg.node import TrueTrigger, FalseTrigger, InputTrigger
from iotile.sg import DataStreamSelector
from iotile.sg.parser.stream_allocator import StreamAllocator


class RemoveDeadCodePass(object):
    """Remove nodes whose operation is discarded and has no side effects.

    An operation has no side effects if it is not a call_rpc operation
    (currently) since no one outside the sensorgraph engine knows if
    the operation was run or not.  It's output is discarded if its stream
    is autogenerated by the compiler and it is not connected to anything.

    Args:
        sensor_graph (SensorGraph): The sensor graph to run
            the optimization pass on
    """

    def __init__(self):
        pass

    def run(self, sensor_graph, model):
        """Run this optimization pass on the sensor graph

        If necessary, information on the device model being targeted
        can be found in the associated model argument.

        Args:
            sensor_graph (SensorGraph): The sensor graph to optimize
            model (DeviceModel): The device model we're using
        """

        # We can only eliminate a node if the following checks are true
        # 1. It has no other nodes connected to it
        # 2. Its stream is not an output of the entire sensor graph
        # 3. Its stream is autogenerated by the compiler
        # 4. Its operation has no side effects
        # 5. Its stream is not buffered so the value will not be accessible

        for node, inputs, outputs in sensor_graph.iterate_bfs():
            can_remove = False

            # Check 1
            if len(outputs) != 0:
                continue

            # Check 2
            if sensor_graph.is_output(node.stream):
                continue

            # Check 3
            if node.stream.stream_id < StreamAllocator.StartingID:
                continue

            # Check 4
            if node.func_name == u'call_rpc':
                continue

            # Check 5
            if node.stream.buffered:
                # FIXME: Add a warning here if the stream is buffered since
                # its weird for the user to be saving useless data to flash
                continue

            # If all of the checks above have passed, we have found a useless
            # node, let's remove it and return True so we run the pass again
            # and look for additional nodes that are now made useles because
            # of the removal of this one.
            for input_node in inputs:
                input_node.outputs.remove(node)

            if node in sensor_graph.roots:
                sensor_graph.roots.remove(node)

            sensor_graph.nodes.remove(node)

            # FIXME: Check if we need to destroy any walkers here

            return True

        return False



"""Copy stream statement."""

from builtins import str
from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from ...stream import DataStream


@python_2_unicode_compatible
class CopyStatement(SensorGraphStatement):
    """Copy one stream into another

    The form of the statement should be
    copy [all | count | average] [stream | number] => <output stream>

    If stream is passed, it is the stream that is copied when the current
    scope's trigger fires.  Otherwise the trigger value is copied.

    You can also copy a fixed constant value by passing an explicit number.

    If all or count are specified all readings are copied or the count
    of the number of readings is copied instead of just the latest reading.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        location (LocationInfo): A namedtuple with information on the line this
            statement was generated from so that we can log appropriate error
            messages.
    """

    def __init__(self, parsed, location=None):
        self.all = 'modifier' in parsed and parsed['modifier'] == u'all'
        self.count = 'modifier' in parsed and parsed['modifier'] == u'count'
        self.average = 'modifier' in parsed and parsed['modifier'] == u'average'

        self.explicit_input = None
        if 'explicit_input' in parsed:
            self.explicit_input = parsed['explicit_input'][0]

        self.constant_input = None
        if 'constant_input' in parsed:
            self.constant_input = parsed['constant_input']

        self.output = parsed['output'][0]

        super(CopyStatement, self).__init__([], location)

    def __str__(self):
        op = u'copy'

        if self.all:
            op = u'copy all'
        elif self.average:
            op = u'copy average'
        elif self.count:
            op = u'copy count'

        input_stream = u""
        if self.explicit_input:
            input_stream = u' ' + str(self.explicit_input)
        elif self.constant_input is not None:
            input_stream = u' ' + str(self.constant_input)

        return u'{}{} => {};'.format(op, input_stream, self.output)

    def execute(self, sensor_graph, scope_stack):
        """Execute this statement on the sensor_graph given the current scope tree.

        This adds a single node to the sensor graph with either the
        copy_latest_a, copy_all_a or average_a function as is processing function.

        If there is an explicit stream passed, that is used as input a with the
        current scope's trigger as input b, otherwise the current scope's trigger
        is used as input a.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]
        alloc = parent.allocator

        trigger_stream, trigger_cond = parent.trigger_chain()

        op = 'copy_latest_a'
        if self.all:
            op = 'copy_all_a'
        elif self.average:
            op = 'average_a'
        elif self.count:
            op = 'copy_count_a'

        if self.explicit_input:
            sensor_graph.add_node(u"({} always && {} {}) => {} using {}".format(self.explicit_input, trigger_stream, trigger_cond, self.output, op))
        elif self.constant_input is not None:
            const_stream = alloc.allocate_stream(DataStream.ConstantType, attach=True)

            sensor_graph.add_node(u"({} always && {} {}) => {} using {}".format(const_stream, trigger_stream, trigger_cond, self.output, op))
            sensor_graph.add_constant(const_stream, self.constant_input)
        else:
            sensor_graph.add_node(u"({} {}) => {} using {}".format(trigger_stream, trigger_cond, self.output, op))



import pytest

from iotile.sg.model import DeviceModel
from iotile.sg.sensor_log import SensorLog
from iotile.sg import DataStreamSelector, DataStream, StreamEmptyError
from iotile.core.hw.reports import IOTileReading


def test_counter_walker():
    """Make sure counter walkers work correctly."""

    model = DeviceModel()
    log = SensorLog(model=model)

    walk = log.create_walker(DataStreamSelector.FromString('counter 1'))
    stream = DataStream.FromString('counter 1')
    reading = IOTileReading(0, 1, 1)

    assert walk.count() == 0
    log.push(stream, IOTileReading(0, 1, 1))
    assert walk.count() == 1
    assert walk.peek().value == 1

    log.push(stream, IOTileReading(0, 1, 3))
    assert walk.count() == 2
    assert walk.peek().value == 3

    val = walk.pop()
    assert walk.count() == 1
    assert val.value == 3

    val = walk.pop()
    assert walk.count() == 0
    assert val.value == 3

    with pytest.raises(StreamEmptyError):
        walk.pop()


def test_unbuffered_walker():
    """Make sure unbuffered walkers hold only 1 reading."""

    model = DeviceModel()
    log = SensorLog(model=model)

    walk = log.create_walker(DataStreamSelector.FromString('unbuffered 1'))
    stream = DataStream.FromString('unbuffered 1')

    assert walk.count() == 0
    log.push(stream, IOTileReading(0, 1, 1))
    assert walk.count() == 1
    assert walk.peek().value == 1

    log.push(stream, IOTileReading(0, 1, 3))
    assert walk.count() == 1
    assert walk.peek().value == 3

    val = walk.pop()
    assert walk.count() == 0
    assert val.value == 3

    with pytest.raises(StreamEmptyError):
        walk.pop()


def test_constant_walker():
    """Make sure constant walkers can be read any number of times."""

    model = DeviceModel()
    log = SensorLog(model=model)

    walk = log.create_walker(DataStreamSelector.FromString('constant 1'))
    stream = DataStream.FromString('constant 1')
    reading = IOTileReading(0, 1, 1)

    log.push(stream, reading)
    assert walk.count() == 0xFFFFFFFF

    log.push(stream, reading)
    assert walk.count() == 0xFFFFFFFF

    val = walk.pop()
    assert walk.count() == 0xFFFFFFFF

    val = walk.pop()
    assert walk.count() == 0xFFFFFFFF


def test_storage_walker():
    """Make sure the storage walker works."""

    model = DeviceModel()
    log = SensorLog(model=model)


    walk = log.create_walker(DataStreamSelector.FromString('buffered 1'))
    stream = DataStream.FromString('buffered 1')
    
    storage_size = model.get('max_storage_buffer')
    erase_size = model.get('buffer_erase_size')

    for i in range(0, storage_size):
        reading = IOTileReading(0, stream.encode(), i)
        log.push(stream, reading)

        assert walk.count() == (i + 1)

    # Make sure the overwrite happens correctly
    old_count = walk.count()
    reading = IOTileReading(0, stream.encode(), storage_size)
    log.push(stream, reading)
    assert walk.count() == (old_count - erase_size + 1)


def test_storage_streaming_walkers():
    """Make sure the storage and streaming walkers work simultaneously."""

    model = DeviceModel()
    log = SensorLog(model=model)


    storage_walk = log.create_walker(DataStreamSelector.FromString('buffered 1'))
    output_walk = log.create_walker(DataStreamSelector.FromString('output 1'))
    storage1 = DataStream.FromString('buffered 1')
    storage2 = DataStream.FromString('buffered 2')
    output1 = DataStream.FromString('output 1')
    output2 = DataStream.FromString('output 2')

    assert storage_walk.offset == 0
    assert output_walk.offset == 0

    for i in range(0, 5000):
        reading = IOTileReading(0, 0, i)
        log.push(storage1, reading)
        log.push(storage2, reading)
        log.push(output1, reading)
        log.push(output2, reading)

        assert storage_walk.count() == (i + 1)
        assert output_walk.count() == (i + 1)
        assert storage_walk.offset == 0
        assert output_walk.offset == 0

    for i in range(0, 5000):
        reading1 = storage_walk.pop()
        reading2 = output_walk.pop()

        assert reading1.value == i
        assert reading1.stream == storage1.encode()
        assert reading2.value == i
        assert reading2.stream == output1.encode()
        assert storage_walk.offset == (2 * i) + 1
        assert output_walk.offset == (2* i) + 1

    log.clear()

    assert storage_walk.count() == 0
    assert storage_walk.offset == 0

    assert output_walk.count() == 0
    assert output_walk.offset == 0



"""Tests for DataStream objects."""

from builtins import str
from iotile.sg import DataStream, DataStreamSelector


def test_stream_type_parsing():
    """Make sure we can parse each type of stream."""

    # Make sure parsing stream type works
    stream = DataStream.FromString('buffered 1')
    assert stream.stream_type == stream.BufferedType
    stream = DataStream.FromString(u'buffered 1')
    assert stream.stream_type == stream.BufferedType

    stream = DataStream.FromString('unbuffered 1')
    assert stream.stream_type == stream.UnbufferedType
    stream = DataStream.FromString(u'unbuffered 1')
    assert stream.stream_type == stream.UnbufferedType

    stream = DataStream.FromString('counter 1')
    assert stream.stream_type == stream.CounterType
    stream = DataStream.FromString(u'counter 1')
    assert stream.stream_type == stream.CounterType

    stream = DataStream.FromString('constant 1')
    assert stream.stream_type == stream.ConstantType
    stream = DataStream.FromString(u'constant 1')
    assert stream.stream_type == stream.ConstantType

    stream = DataStream.FromString('output 1')
    assert stream.stream_type == stream.OutputType
    stream = DataStream.FromString(u'output 1')
    assert stream.stream_type == stream.OutputType


def test_stream_id_parsing():
    """Make sure we can parse stream ids."""

    stream = DataStream.FromString('buffered 1')
    assert stream.stream_id == 1

    stream = DataStream.FromString('buffered 0x100')
    assert stream.stream_id == 0x100

    stream = DataStream.FromString(u'buffered 1')
    assert stream.stream_id == 1

    stream = DataStream.FromString(u'buffered 0x100')
    assert stream.stream_id == 0x100


def test_system_parsing():
    """Make sure we can parse the system prefix."""

    stream = DataStream.FromString('buffered 1')
    assert stream.system is False
    stream = DataStream.FromString(u'buffered 1')
    assert stream.system is False

    stream = DataStream.FromString('system buffered 1')
    assert stream.system is True
    stream = DataStream.FromString(u'system buffered 1')
    assert stream.system is True


def test_stringification():
    """Make sure we can stringify DataStream objects."""

    stream1 = DataStream.FromString('system buffered 1')
    stream2 = DataStream.FromString('buffered 0xF')

    assert str(stream1) == str('system buffered 1')
    assert str(stream2) == str('buffered 15')


def test_selector_parsing():
    """Make sure we can parse DataStreamSelector strings."""

    # Make sure parsing stream type works
    stream = DataStreamSelector.FromString('buffered 1')
    assert stream.match_type == DataStream.BufferedType
    stream = DataStreamSelector.FromString(u'buffered 1')
    assert stream.match_type == DataStream.BufferedType

    stream = DataStreamSelector.FromString('unbuffered 1')
    assert stream.match_type == DataStream.UnbufferedType
    stream = DataStreamSelector.FromString(u'unbuffered 1')
    assert stream.match_type == DataStream.UnbufferedType

    stream = DataStreamSelector.FromString('counter 1')
    assert stream.match_type == DataStream.CounterType
    stream = DataStreamSelector.FromString(u'counter 1')
    assert stream.match_type == DataStream.CounterType

    stream = DataStreamSelector.FromString('constant 1')
    assert stream.match_type == DataStream.ConstantType
    stream = DataStreamSelector.FromString(u'constant 1')
    assert stream.match_type == DataStream.ConstantType

    stream = DataStreamSelector.FromString('output 1')
    assert stream.match_type == DataStream.OutputType
    stream = DataStreamSelector.FromString(u'output 1')
    assert stream.match_type == DataStream.OutputType


def test_stream_selector_id_parsing():
    """Make sure we can parse stream ids."""

    stream = DataStreamSelector.FromString('buffered 1')
    assert stream.match_id == 1
    assert stream.match_spec == DataStreamSelector.MatchUserOnly

    stream = DataStreamSelector.FromString('buffered 0x100')
    assert stream.match_id == 0x100
    assert stream.match_spec == DataStreamSelector.MatchUserOnly

    stream = DataStreamSelector.FromString(u'buffered 1')
    assert stream.match_id == 1
    assert stream.match_spec == DataStreamSelector.MatchUserOnly

    stream = DataStreamSelector.FromString(u'buffered 0x100')
    assert stream.match_id == 0x100
    assert stream.match_spec == DataStreamSelector.MatchUserOnly

    stream = DataStreamSelector.FromString(u'system buffered 0x100')
    assert stream.match_id == 0x100
    assert stream.match_spec == DataStreamSelector.MatchSystemOnly

    stream = DataStreamSelector.FromString(u'all buffered')
    assert stream.match_id is None
    assert stream.match_spec == DataStreamSelector.MatchUserAndBreaks

    stream = DataStreamSelector.FromString(u'all user buffered')
    assert stream.match_id is None
    assert stream.match_spec == DataStreamSelector.MatchUserOnly

    stream = DataStreamSelector.FromString(u'all combined buffered')
    assert stream.match_id is None
    assert stream.match_spec == DataStreamSelector.MatchCombined

    stream = DataStreamSelector.FromString(u'all system buffered')
    assert stream.match_id is None
    assert stream.match_spec == DataStreamSelector.MatchSystemOnly


def test_matching():
    """Test selector stream matching."""

    sel = DataStreamSelector.FromString(u'all system buffered')
    assert sel.matches(DataStream.FromString('system buffered 1'))
    assert not sel.matches(DataStream.FromString('buffered 1'))
    assert not sel.matches(DataStream.FromString('counter 1'))

    sel = DataStreamSelector.FromString(u'all user outputs')
    assert sel.matches(DataStream.FromString('output 1'))
    assert not sel.matches(DataStream.FromString('system output 1'))
    assert not sel.matches(DataStream.FromString('counter 1'))

    sel = DataStreamSelector.FromString(u'all combined outputs')
    assert sel.matches(DataStream.FromString('output 1'))
    assert sel.matches(DataStream.FromString('system output 1'))
    assert not sel.matches(DataStream.FromString('counter 1'))

    sel = DataStreamSelector.FromString(u'all outputs')
    assert sel.matches(DataStream.FromString('output 1'))
    assert sel.matches(DataStream.FromString('system output 1024'))
    assert not sel.matches(DataStream.FromString('system output 1'))
    assert not sel.matches(DataStream.FromString('counter 1'))


def test_encoding():
    """Test data stream and selector encoding."""

    sel = DataStreamSelector.FromString(u'all system output')
    assert sel.encode() == 0x5FFF

    sel = DataStreamSelector.FromString(u'all user output')
    assert sel.encode() == 0x57FF

    sel = DataStreamSelector.FromString(u'all output')
    assert sel.encode() == 0xD7FF

    sel = DataStreamSelector.FromString(u'all combined output')
    assert sel.encode() == 0xDFFF

    stream = DataStream.FromString('output 1')
    assert stream.encode() == 0x5001

    stream = DataStream.FromString('unbuffered 10')
    assert stream.encode() == 0x100a


def test_selector_from_encoded():
    """Make sure we can create a selector from an encoded value."""

    sel = DataStreamSelector.FromEncoded(0x5FFF)
    assert str(sel) == 'all system outputs'

    sel = DataStreamSelector.FromEncoded(0xD7FF)
    assert str(sel) == 'all outputs'

    sel = DataStreamSelector.FromEncoded(0x100a)
    assert str(sel) == 'unbuffered 10'

    assert str(DataStreamSelector.FromEncoded(DataStreamSelector.FromString('all combined output').encode())) == 'all combined outputs'


def test_buffered_pluralization():
    """Make sure we don't incorrectly pluralize buffered streams."""

    sel = DataStreamSelector.FromString('all buffered')
    assert str(sel) == 'all buffered'



"""A machine parsable ascii representation of a sensor graph.

This format is directly loadable by the sensor_graph proxy using
the load_from_file method.  It closely mirrors the snippet format
with the arguments enclosed in brackets for easier parsing.
"""

from iotile.core.utilities.command_file import CommandFile


def format_ascii(sensor_graph):
    """Format this sensor graph as a loadable ascii file format.

    This includes commands to reset and clear previously stored
    sensor graphs.

    NB. This format does not include any required configuration
    variables that were specified in this sensor graph, so you
    should also output tha information separately in, e.g.
    the config format.

    Args:
        sensor_graph (SensorGraph): the sensor graph that we want to format

    Returns:
        str: The ascii output lines concatenated as a single string
    """

    cmdfile = CommandFile("Sensor Graph", "1.0")

    # Clear any old sensor graph
    cmdfile.add("set_online", False)
    cmdfile.add("clear")
    cmdfile.add("reset")

    # Load in the nodes
    for node in sensor_graph.dump_nodes():
        cmdfile.add('add_node', node)

    # Load in the streamers
    for streamer in sensor_graph.streamers:
        other = 0xFF
        if streamer.with_other is not None:
            other = streamer.with_other

        args = [streamer.walker.selector, streamer.dest, streamer.automatic, streamer.format, streamer.report_type, other]
        cmdfile.add('add_streamer', *args)

    # Load all the constants
    for stream, value in sensor_graph.constant_database.items():
        cmdfile.add("push_reading", stream, value)

    # Persist the sensor graph
    cmdfile.add("persist")
    cmdfile.add("set_online", True)

    return cmdfile.dump()



import logging
from ...node import InputTrigger
from ...stream import DataStream
from ...exceptions import SensorGraphSemanticError
from .scope import Scope


class GatedClockScope(Scope):
    """A scope that will gate all requested clocks with a latch.

    Args:
        sensor_graph (SensorGraph): The sensor graph we are working on.
        scope_stack (list(Scope)): The stack of already allocated scopes.
        input_latch ((DataStream, InputTrigger)): The input stream and condition
            that should be used to gate clocks passed through this scope. The
            stream must already be attached.
    """

    def __init__(self, sensor_graph, scope_stack, input_latch):
        parent = scope_stack[-1]
        alloc = parent.allocator
        sensor_graph = parent.sensor_graph

        super(GatedClockScope, self).__init__(u"Gated Clock Scope", sensor_graph, alloc, parent)

        stream = alloc.allocate_stream(DataStream.ConstantType)
        sensor_graph.add_node(u'({} always) => {} using copy_latest_a'.format(input_latch[0], stream))
        self.latch_stream = stream
        self.latch_trigger = input_latch[1]
        self.clock_cache = {}
        self.logger = logging.getLogger(__name__)

        self.logger.debug("Allocating GatedClockScope on latch stream %s with condition %s", input_latch[0], input_latch[1])

    def _classify_clock(self, interval, basis):
        if basis == 'system':
            if interval % 10 == 0:
                return 'standard'

            return 'fast'
        elif basis == 'tick_1':
            return 'tick_1'
        elif basis == 'tick_2':
            return 'tick_2'

        raise SensorGraphSemanticError("Unknown clock basis in GatedClockScope", scope=self.name, basis=basis, interval=interval)

    def clock(self, interval, basis):
        """Return a NodeInput tuple for triggering an event every interval.

        We request each distinct type of clock at most once and combine it with our
        latch stream each time it is requested.

        Args:
            interval (int): The interval (in seconds) at which this input should
                trigger.
        """

        cache_name = self._classify_clock(interval, basis)
        cache_data = self.clock_cache.get(cache_name)

        if cache_data is None:
            parent_stream, trigger = self.parent.clock(interval, basis)

            if trigger.use_count is False:
                raise SensorGraphSemanticError("Unsupported clock trigger in GatedClockScope", trigger=trigger)
            elif interval % trigger.reference != 0:
                raise SensorGraphSemanticError("Unsupported trigger ratio in GatedClockScope", trigger=trigger, interval=interval)

            ratio = interval // trigger.reference

            stream = self.allocator.allocate_stream(DataStream.CounterType)
            latch_stream = self.allocator.attach_stream(self.latch_stream)

            self.sensor_graph.add_node(u'({} always && {} {}) => {} using copy_latest_a'.format(parent_stream, latch_stream, self.latch_trigger, stream))
            self.clock_cache[cache_name] = (stream, ratio)
        else:
            stream, ratio = cache_data

        if interval % ratio != 0:
            raise SensorGraphSemanticError("Unsupported trigger ratio in GatedClockScope", ratio=ratio, interval=interval)

        count = interval // ratio

        clock_stream = self.allocator.attach_stream(stream)
        return clock_stream, InputTrigger(u'count', '>=', count)



import pytest

from iotile.sg import SlotIdentifier
from iotile.core.exceptions import ArgumentError

def test_slotid_parsing():
    """Make sure we can parse slot identifiers correctly."""

    slot_id = SlotIdentifier.FromString('slot 1')
    assert slot_id.address == 11

    slot_id = SlotIdentifier.FromString(u'slot 1')
    assert slot_id.address == 11

    slot_id = SlotIdentifier.FromString(u'slot 0x1')
    assert slot_id.address == 11

    slot_id = SlotIdentifier.FromString('controller')
    assert slot_id.address == 8

    slot_id = SlotIdentifier.FromString(u'controller')
    assert slot_id.address == 8

    with pytest.raises(ArgumentError):
        SlotIdentifier.FromString(u'asdf')

    with pytest.raises(ArgumentError):
        SlotIdentifier.FromString(u'controller 15')

    with pytest.raises(ArgumentError):
        SlotIdentifier.FromString(u'slot ab')

    slot1 = SlotIdentifier.FromString('slot 1')
    slot2 = SlotIdentifier.FromString('slot 1')
    con = SlotIdentifier.FromString('controller')

    assert slot1 == slot2
    assert not con == slot1
    assert con != slot1


def test_slotid_binary_parsing():
    """Make sure we can generate and parse binary descriptors."""

    slot_id = SlotIdentifier.FromString('slot 1')
    assert SlotIdentifier.FromEncoded(slot_id.encode()) == slot_id

    con = SlotIdentifier.FromString('controller')
    assert SlotIdentifier.FromEncoded(con.encode()) == con



"""Remove unnecessary nodes with the copy_latest_a function.

These nodes are only necessary if there are not enough output links
on the input node to accomodate all outputs or if there is a
triggering condition on an input other than intput A that triggers
the node to copy.
"""
import logging
from iotile.core.exceptions import ArgumentError
from iotile.sg.node import TrueTrigger, FalseTrigger, InputTrigger
from iotile.sg import DataStreamSelector
import copy


class RemoveCopyLatestPass(object):
    """Run the remove copy latest optimization pass.

    Args:
        sensor_graph (SensorGraph): The sensor graph to run
            the optimization pass on
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def run(self, sensor_graph, model):
        """Run this optimization pass on the sensor graph

        If necessary, information on the device model being targeted
        can be found in the associated model argument.

        Args:
            sensor_graph (SensorGraph): The sensor graph to optimize
            model (DeviceModel): The device model we're using
        """

        # Try to remove a node whose operation is copy_latest_a
        # It inherently does nothing, so it can be removed if:
        # 1. There is only one input triggering it
        # 2. The input is from another node (i.e. not a global sg input)
        # 3. The type of the input is the same as the type of the output
        # 4. There are enough output links in the input node to
        #    handle all of the downstream connections.  Note that we will
        #    reclaim one free output when we do the replacement.
        #    If there are multiple nodes that write to this input, all
        #    must have an available output.
        # 5. The stream is not a sensor graph output
        # 6. The trigger conditions for both inputs and outputs are either
        #    count based or always or value based with the same value.
        #
        # For each node, check if 1-6 are valid so we can remove it

        found_node = None
        found_inputs = None
        found_outputs = None

        for node, inputs, outputs in sensor_graph.iterate_bfs():
            if node.func_name != u'copy_latest_a':
                continue

            # Check 1
            if node.num_inputs != 1:
                continue

            # Check 2
            if len(inputs) == 0:
                continue

            can_combine = True
            for curr_input in inputs:
                # Check 3
                if curr_input.stream.stream_type != node.stream.stream_type:
                    can_combine = False
                    break

                # Check 4 (keep in mind we free up one output when we do the swap)
                if (curr_input.free_outputs + 1) < len(outputs):
                    can_combine = False
                    break

            if not can_combine:
                continue

            # Check 5
            if sensor_graph.is_output(node.stream):
                continue

            # Check 6
            for out in outputs:
                i = out.find_input(node.stream)
                _, trigger = out.inputs[i]

                # We can't merge things that could result in producing
                # multiple readings at a time since then combining the
                # trigger might change what number of outputs are
                # produced since:
                # trigger every 600 copy_latest then trigger every 1 copy_all
                # would produce one reading every 600 ticks.
                #
                # but:
                # trigger every 1 copy_latest then trigger every 600 copy_all
                # would produce 600 readings one every 600 ticks.
                if out.func_name == u'copy_all_a':
                    can_combine = False
                    break

                if not self._can_combine(node.inputs[0][1], trigger):
                    can_combine = False
                    break

            if not can_combine:
                continue

            found_node = node
            found_inputs = inputs
            found_outputs = outputs
            break

        if found_node is None:
            return False

        sensor_graph.nodes.remove(found_node)

        for input_node in found_inputs:
            input_node.outputs.remove(found_node)

        for output in found_outputs:
            i = output.find_input(found_node.stream)
            old_walker, old_trigger = output.inputs[i]

            new_trigger = self._try_combine(found_node.inputs[0][1], old_trigger)
            new_walker = sensor_graph.sensor_log.create_walker(DataStreamSelector.FromString(str(found_node.inputs[0][0].selector)))
            sensor_graph.sensor_log.destroy_walker(old_walker)

            output.inputs[i] = (new_walker, new_trigger)

            for input_node in found_inputs:
                input_node.connect_output(output)

        if found_node in sensor_graph.roots:
            sensor_graph.roots.remove(found_node)

        sensor_graph.sensor_log.destroy_walker(found_node.inputs[0][0])
        return True

    def _can_combine(self, trigger1, trigger2):
        """Check if we can combine two triggers together.

        Return:
            bool
        """

        try:
            self._try_combine(trigger1, trigger2)
            return True
        except ArgumentError:
            pass

        return False

    def _try_combine(self, trigger1, trigger2):
        """Try to combine two triggers together.

        This function assumes a node structure of
            trigger1 -> copy_latest -> trigger2
        and tries to find a single trigger that can replace this
        as:
            trigger3
        with no copy latest

        Throws:
            ArgumentError: there is no way to combine these two
                triggers into one

        Returns:
            object: The combined trigger if possible
        """

        if isinstance(trigger1, FalseTrigger) or isinstance(trigger2, FalseTrigger):
            return FalseTrigger()

        if isinstance(trigger1, TrueTrigger):
            return copy.copy(trigger2)

        if isinstance(trigger2, TrueTrigger):
            return copy.copy(trigger1)

        # Otherwise these are both InputTrigger instances, which we can only
        # combine in certain cases.

        # If the triggers are count == X and count == Y, we can combine them to
        # count == X*Y
        #
        # If the triggers are count >= X and count == 1 then this is the same
        # as just count >= X since every time that triggers the next node
        # will trigger immediately and consume everything down to 0.
        if trigger1.use_count and trigger2.use_count:
            if trigger1.comp_string == u'==' and trigger2.comp_string == u'==':
                combined_ref = trigger1.reference * trigger2.reference

                # Make sure we won't overflow
                if combined_ref <= 0xFFFFFFFF:
                    return InputTrigger(u'count', u'==', combined_ref)
            elif trigger1.comp_string in [u">=", u"<=", u"<", u">"] and trigger2.comp_string == u"==" and trigger2.reference == 1:
                return InputTrigger(u'count', trigger1.comp_string, trigger1.reference)

        elif trigger1.use_count is False and trigger2.use_count is False:
            # If these are value triggers then we can combine them if the
            # value is the same
            if trigger1.comp_string == u"==" and trigger2.comp_string == u"==" and trigger1.reference == trigger2.reference:
                return InputTrigger(u"value", trigger1.comp_string, trigger1.reference)

        raise ArgumentError("Cannot combine triggers")



"""On blocks trigger functions when an event occurs on a stream."""

from builtins import str
from collections import namedtuple
from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from ..scopes import TriggerScope
from ...node import InputTrigger, TrueTrigger
from ... import DataStream
from iotile.core.exceptions import ArgumentError

TriggerDefinition = namedtuple("TriggerDefinition", ['named_event', 'explicit_stream', 'explicit_trigger'])


@python_2_unicode_compatible
class OnBlock(SensorGraphStatement):
    """A block of statements that should run every time an event occurs

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        children(list(SensorGraphStatement)): The statements that are
            part of this on block.
        location (LocationInfo): A namedtuple with information on the line this
            statement was generated from so that we can log appropriate error
            messages.
    """

    def __init__(self, parsed, children, location=None):
        trigger_a = parsed[0]

        self.trigger_a = self._parse_trigger(trigger_a)

        if len(parsed) > 1:
            self.combiner = parsed[1]
            self.trigger_b = self._parse_trigger(parsed[2])
        else:
            self.trigger_b = None
            self.combiner = None

        super(OnBlock, self).__init__(children, location)

    def _format_trigger(self, trigger_def):
        if trigger_def.explicit_stream is not None:
            return trigger_def.explicit_trigger.format_trigger(trigger_def.explicit_stream)

        return u"{}".format(trigger_def.named_event)

    def _convert_trigger(self, trigger_def, parent):
        """Convert a TriggerDefinition into a stream, trigger pair."""

        if trigger_def.explicit_stream is None:
            stream = parent.resolve_identifier(trigger_def.named_event, DataStream)
            trigger = TrueTrigger()
        else:
            stream = trigger_def.explicit_stream
            trigger = trigger_def.explicit_trigger

        return (stream, trigger)

    def _parse_trigger(self, trigger_clause):
        """Parse a named event or explicit stream trigger into a TriggerDefinition."""

        cond = trigger_clause[0]

        named_event = None
        explicit_stream = None
        explicit_trigger = None

        # Identifier parse tree is Group(Identifier)
        if cond.getName() == 'identifier':
            named_event = cond[0]
        elif cond.getName() == 'stream_trigger':
            trigger_type = cond[0]
            stream = cond[1]
            oper = cond[2]
            ref = cond[3]

            trigger = InputTrigger(trigger_type, oper, ref)
            explicit_stream = stream
            explicit_trigger = trigger
        elif cond.getName() == 'stream_always':
            stream = cond[0]
            trigger = TrueTrigger()
            explicit_stream = stream
            explicit_trigger = trigger
        else:
            raise ArgumentError("OnBlock created from an invalid ParseResults object", parse_results=trigger_clause)

        return TriggerDefinition(named_event, explicit_stream, explicit_trigger)

    def __str__(self):
        if self.combiner is None:
            return u"on {}".format(self._format_trigger(self.trigger_a))

        return u"on {} {} {}".format(self._format_trigger(self.trigger_a), self.combiner, self._format_trigger(self.trigger_b))

    def execute_before(self, sensor_graph, scope_stack):
        """Execute statement before children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]
        alloc = parent.allocator

        stream_a, trigger_a = self._convert_trigger(self.trigger_a, parent)

        if self.trigger_b is None:
            new_scope = TriggerScope(sensor_graph, scope_stack, (stream_a, trigger_a))
        else:
            stream_b, trigger_b = self._convert_trigger(self.trigger_b, parent)
            trigger_stream = alloc.allocate_stream(DataStream.UnbufferedType)

            if self.combiner == u'and':
                combiner = '&&'
            else:
                combiner = '||'

            sensor_graph.add_node(u"({} {} {} {} {}) => {} using copy_latest_a".format(stream_a, trigger_a, combiner, stream_b, trigger_b, trigger_stream))
            new_scope = TriggerScope(sensor_graph, scope_stack, (trigger_stream, TrueTrigger()))

        scope_stack.append(new_scope)

    def execute_after(self, sensor_graph, scope_stack):
        """Execute statement after children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        scope_stack.pop()



from .remove_copylatest import RemoveCopyLatestPass
from .convert_to_always import ConvertCountOneToAlways
from .downgrade_copyall import ConvertCopyAllToCopyLatest
from .dead_code_elimination import RemoveDeadCodePass
from .remove_constants import RemoveConstantsPass

__all__ = ['RemoveCopyLatestPass', 'ConvertCountOneToAlways', 'ConvertCopyAllToCopyLatest',
           'RemoveDeadCodePass', 'RemoveConstantsPass']



from ..stream import DataStream
from iotile.core.exceptions import ArgumentError


class StreamAllocator(object):
    """Singleton object for allocating and managing streams.

    The StreamAllocator is in charge of allocating streams as needed
    from the pool of available streams for different purposes.  It also
    keeps track of how many inputs are attached to a given stream and splits
    it when there are two many streams attached.

    Args:
        sensor_graph (SensorGraph): The sensor graph that we allocating
            streams for.
        model (DeviceModel): The device model we are allocating for.
            This is required so that the StreamAllocator knows when to
            split a stream into two since we're out of outputs.
    """

    StartingID = 0x400

    def __init__(self, sensor_graph, model):
        self.sensor_graph = sensor_graph
        self.model = model

        self._allocated_streams = {}
        self._next_id = {}

    def allocate_stream(self, stream_type, stream_id=None, previous=None, attach=False):
        """Allocate a new stream of the given type.

        The stream is allocated with an incremental ID starting at
        StreamAllocator.StartingID.  The returned data stream can always
        be used to to attach a NodeInput to this stream, however the
        attach_stream() function should always be called first since this
        stream's output may need to be split and a logically equivalent
        stream used instead to satisfy a device specific constraint on the
        maximum number of outputs attached to a given stream.

        You can call allocate_stream on the same stream multiple times without
        issue.  Subsequent calls to allocate_stream are noops.

        Args:
            stream_type (int): A stream type specified in the DataStream class
                like DataStream.ConstantType
            stream_id (int): The ID we would like to use for this stream, if
                this is not specified, an ID is automatically allocated.
            previous (DataStream): If this stream was automatically derived from
                another stream, this parameter should be a link to the old
                stream.
            attach (bool): Call attach_stream immediately before returning.  Convenience
                routine for streams that should immediately be attached to something.

        Returns:
            DataStream: The allocated data stream.
        """

        if stream_type not in DataStream.TypeToString:
            raise ArgumentError("Unknown stream type in allocate_stream", stream_type=stream_type)

        if stream_id is not None and stream_id >= StreamAllocator.StartingID:
            raise ArgumentError("Attempted to explicitly allocate a stream id in the internally managed id range", stream_id=stream_id, started_id=StreamAllocator.StartingID)

        # If the stream id is not explicitly given, we need to manage and track it
        # from our autoallocate range
        if stream_id is None:
            if stream_type not in self._next_id:
                self._next_id[stream_type] = StreamAllocator.StartingID

            stream_id = self._next_id[stream_type]
            self._next_id[stream_type] += 1

        # Keep track of how many downstream nodes are attached to this stream so
        # that we know when we need to split it into two.
        stream = DataStream(stream_type, stream_id)

        if stream not in self._allocated_streams:
            self._allocated_streams[stream] = (stream, 0, previous)

        if attach:
            stream = self.attach_stream(stream)

        return stream

    def attach_stream(self, stream):
        """Notify that we would like to attach a node input to this stream.

        The return value from this function is the DataStream that should be attached
        to since this function may internally allocate a new SGNode that copies the
        stream if there is no space in the output list to hold another input.

        This function should be called once for every node input before allocated a new
        sensor graph node that attaches to a stream that is managed by the StreamAllocator.

        Args:
            stream (DataStream): The stream (originally returned from allocate_stream)
                that we want to attach to.

        Returns:
            Datastream: A data stream, possible the same as stream, that should be attached
                to a node input.
        """

        curr_stream, count, prev = self._allocated_streams[stream]

        # Check if we need to split this stream and allocate a new one
        if count == (self.model.get(u'max_node_outputs') - 1):
            new_stream = self.allocate_stream(curr_stream.stream_type, previous=curr_stream)
            copy_desc = u"({} always) => {} using copy_all_a".format(curr_stream, new_stream)

            self.sensor_graph.add_node(copy_desc)
            self._allocated_streams[stream] = (new_stream, 1, curr_stream)

            # If we are splitting a constant stream, make sure we also duplicate the initialization value
            # FIXME: If there is no default value for the stream, that is probably a warning since all constant
            #        streams should be initialized with a value.
            if curr_stream.stream_type == DataStream.ConstantType and curr_stream in self.sensor_graph.constant_database:
                self.sensor_graph.add_constant(new_stream, self.sensor_graph.constant_database[curr_stream])

            return new_stream

        self._allocated_streams[stream] = (curr_stream, count + 1, prev)
        return curr_stream



"""Remove unnecessary constants."""

import logging
from iotile.sg import DataStream


class RemoveConstantsPass(object):
    """Run the remove constants optimization pass.

    Sometimes we remove nodes that would otherwise have a constant
    input and thus there are constants remaining in the sensor_graph
    added via `add_constant` that no longer exist in the underlying
    graph structure.  This optimization pass prunes those constants
    out.

    It is important because setting a nonexistent constant in some
    embedded sensorgraph engines is not allowed since storage for
    constant streams is entirely contained within the nodes that
    use them for inputs.

    Args:
        sensor_graph (SensorGraph): The sensor graph to run
            the optimization pass on
    """

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def run(self, sensor_graph, model):
        inuse_constants = set()

        for node, _inputs, _outputs in sensor_graph.iterate_bfs():
            for input_walker, _trigger in node.inputs:
                if input_walker.selector is None:
                    continue

                if input_walker.selector.match_type == DataStream.ConstantType:
                    inuse_constants.add(input_walker.selector.as_stream())

        all_constants = set(sensor_graph.constant_database.keys())

        to_remove = all_constants - inuse_constants
        self.logger.debug("RemoveConstantsPass removing %d unused constants", len(to_remove))

        for val in to_remove:
            del sensor_graph.constant_database[val]



"""Test to make sure we can create and use SensorGraph objects."""

from iotile.sg import SensorGraph, DeviceModel, SensorLog, DataStream, SlotIdentifier
from iotile.sg.known_constants import config_fast_tick_secs
from iotile.core.hw.reports import IOTileReading


def test_basic_sensorgraph():
    """Make sure we can parse, load and run a basic sensor graph."""

    model = DeviceModel()
    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)

    sg.add_node('(input 1 always && input 2 when count >= 1) => unbuffered 1 using copy_all_a')
    sg.process_input(DataStream.FromString('input 1'), IOTileReading(0, 1, 1), rpc_executor=None)
    sg.process_input(DataStream.FromString('input 2'), IOTileReading(0, 1, 1), rpc_executor=None)

    assert sg.sensor_log.inspect_last(DataStream.FromString('unbuffered 1')).value == 1


def test_usertick():
    """Make sure we properly can set the user tick input."""

    model = DeviceModel()
    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)

    assert sg.get_tick('fast') == 0

    sg.add_config(SlotIdentifier.FromString('controller'), config_fast_tick_secs, 'uint32_t', 1)
    assert sg.get_tick('fast') == 1


def test_iteration():
    """Make sure we can iterate over the graph."""

    model = DeviceModel()
    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)

    sg.add_node('(input 1 always && input 2 when count >= 1) => unbuffered 1 using copy_all_a')
    sg.add_node('(input 1 always && input 3 when count >= 1) => unbuffered 2 using copy_all_a')
    sg.add_node('(unbuffered 2 always && unbuffered 1 always) => unbuffered 3 using copy_all_a')
    sg.add_node('(unbuffered 1 always) => unbuffered 3 using copy_all_a')

    iterator = sg.iterate_bfs()

    node1, in1, out1 = next(iterator)
    assert str(node1.stream) == u'unbuffered 1'
    assert len(in1) == 0
    assert len(out1) == 2
    assert str(out1[0].stream) == u'unbuffered 3'
    assert str(out1[1].stream) == u'unbuffered 3'

    node1, in1, out1 = next(iterator)
    assert str(node1.stream) == u'unbuffered 2'
    assert len(in1) == 0
    assert len(out1) == 1
    assert str(out1[0].stream) == u'unbuffered 3'

    node1, in1, out1 = next(iterator)
    assert str(node1.stream) == u'unbuffered 3'
    assert len(in1) == 2
    assert len(out1) == 0
    assert str(in1[0].stream) == u'unbuffered 2'
    assert str(in1[1].stream) == u'unbuffered 1'

    node1, in1, out1 = next(iterator)
    assert str(node1.stream) == u'unbuffered 3'
    assert len(in1) == 1
    assert len(out1) == 0
    assert str(in1[0].stream) == u'unbuffered 1'



"""Parsing routines for understanding binary streamer descriptors.

Binary streamer descriptors are the internal binary representation of
a streamer stored in an IOTile device's embedded firmware.
"""

from __future__ import (unicode_literals, absolute_import, print_function)
from builtins import str
import struct
from typedargs.exceptions import ArgumentError
from .slot import SlotIdentifier
from .stream import DataStreamSelector
from .streamer import DataStreamer


def parse_binary_descriptor(bindata):
    """Convert a binary streamer descriptor into a string descriptor.

    Binary streamer descriptors are 20-byte binary structures that encode all
    information needed to create a streamer.  They are used to communicate
    that information to an embedded device in an efficent format.  This
    function exists to turn such a compressed streamer description back into
    an understandable string.

    Args:
        bindata (bytes): The binary streamer descriptor that we want to
            understand.

    Returns:
        str: A string description of the streamer.
    """

    if len(bindata) != 14:
        raise ArgumentError("Invalid length of binary data in streamer descriptor", length=len(bindata), expected=14, data=bindata)

    dest_tile, stream_id, trigger, format_code, type_code = struct.unpack("<8sHBBBx", bindata)

    dest_id = SlotIdentifier.FromEncoded(dest_tile)
    selector = DataStreamSelector.FromEncoded(stream_id)

    format_name = DataStreamer.KnownFormatCodes.get(format_code)
    type_name = DataStreamer.KnownTypeCodes.get(type_code)

    if format_name is None:
        raise ArgumentError("Unknown format code", code=format_code, known_code=DataStreamer.KnownFormatCodes)
    if type_name is None:
        raise ArgumentError("Unknown type code", code=type_code, known_codes=DataStreamer.KnownTypeCodes)

    with_other = None
    if trigger & (1 << 7):
        auto = False
        with_other = trigger & ((1 << 7) - 1)
    elif trigger == 0:
        auto = False
    elif trigger == 1:
        auto = True
    else:
        raise ArgumentError("Unknown trigger type for streamer", trigger_code=trigger)

    manual = "manual " if not auto else ""
    realtime = "realtime " if format_name == 'individual' else ""

    security = ""
    if format_name == 'signedlist_userkey':
        security = "signed "

    to_slot = ""
    if not dest_id.controller:
        to_slot = " to " + str(dest_id)

    with_statement = ""
    if with_other is not None:
        with_statement = " with streamer %d" % with_other

    template = "{manual}{security}{realtime}streamer on {selector}{to_slot}{with_other}"
    return template.format(manual=manual, security=security, realtime=realtime, selector=selector,
                           with_other=with_statement, to_slot=to_slot)



"""A snippet is a series of commands that can be piped into the iotile tool connected to a device."""

from binascii import hexlify

def format_snippet(sensor_graph):
    """Format this sensor graph as iotile command snippets.

    This includes commands to reset and clear previously stored
    sensor graphs.

    Args:
        sensor_graph (SensorGraph): the sensor graph that we want to format
    """

    output = []

    # Clear any old sensor graph
    output.append("disable")
    output.append("clear")
    output.append("reset")

    # Load in the nodes
    for node in sensor_graph.dump_nodes():
        output.append('add_node "{}"'.format(node))

    # Load in the streamers
    for streamer in sensor_graph.streamers:
        line = "add_streamer '{}' '{}' {} {} {}".format(streamer.walker.selector, streamer.dest, streamer.automatic, streamer.format, streamer.report_type)

        if streamer.with_other is not None:
            line += ' --withother {}'.format(streamer.with_other)

        output.append(line)

    # Load all the constants
    for stream, value in sensor_graph.constant_database.items():
        output.append("set_constant '{}' {}".format(stream, value))

    # Persist the sensor graph
    output.append("persist")

    output.append("back")

    # If we have an app tag and version set program them in
    app_tag = sensor_graph.metadata_database.get('app_tag')
    app_version = sensor_graph.metadata_database.get('app_version')

    if app_tag is not None:
        if app_version is None:
            app_version = "0.0"

        output.append("test_interface")
        output.append("set_version app %d --version '%s'" % (app_tag, app_version))
        output.append("back")

    # Load in the config variables if any
    output.append("config_database")
    output.append("clear_variables")

    for slot, conf_vars in sensor_graph.config_database.items():
        for conf_var, conf_def in conf_vars.items():
            conf_type, conf_val = conf_def

            if conf_type == 'binary':
                conf_val = 'hex:' + hexlify(conf_val)
            elif isinstance(conf_val, str):
                conf_val = '"%s"' % conf_val

            output.append("set_variable '{}' {} {} {}".format(slot, conf_var, conf_type, conf_val))

    # Restart the device to load in the new sg
    output.append("back")
    output.append("reset")

    return "\n".join(output) + '\n'



"""A list of trace actions that list what occured during a simulation.

SimulationTrace objects allow one to compare and save the important
details of a simulation while not saving all internal results that
are not directly visible to the external world and could change as
a result of different levels of optimization.
"""

from __future__ import (unicode_literals, absolute_import, print_function)
import json
from iotile.core.hw.reports import IOTileReading
from typedargs.exceptions import ArgumentError
from ..stream import DataStreamSelector, DataStream


class SimulationTrace(list):
    """A trace of all operations that occurred during an SG simulation.

    Traces are creating by attaching watchers to specific stream selectors
    and logging all activity in those streams.

    Args:
        readings (list of IOTileReading): The individual readings that are
            part of this trace.
        selectors (list of DataStreamSelector): The selectors that were used
            to produce this trace.  This information is saved along with the
            raw trace values so that we can compare apples to apples when
            comparing different traces.
    """

    def __init__(self, readings=None, selectors=None):
        if readings is None:
            readings = []
        if selectors is None:
            selectors = []

        self.selectors = selectors
        super(SimulationTrace, self).__init__(readings)

    def save(self, out_path):
        """Save an ascii representation of this simulation trace.

        Args:
            out_path (str): The output path to save this simulation trace.
        """

        out = {
            'selectors': [str(x) for x in self.selectors],
            'trace': [{'stream': str(DataStream.FromEncoded(x.stream)), 'time': x.raw_time, 'value': x.value, 'reading_id': x.reading_id} for x in self]
        }

        with open(out_path, "wb") as outfile:
            json.dump(out, outfile, indent=4)

    @classmethod
    def FromFile(cls, in_path):
        """Load a previously saved ascii representation of this simulation trace.

        Args:
            in_path (str): The path of the input file that we should load.

        Returns:
            SimulationTrace: The loaded trace object.
        """

        with open(in_path, "rb") as infile:
            in_data = json.load(infile)

        if not ('trace', 'selectors') in in_data:
            raise ArgumentError("Invalid trace file format", keys=in_data.keys(), expected=('trace', 'selectors'))

        selectors = [DataStreamSelector.FromString(x) for x in in_data['selectors']]
        readings = [IOTileReading(x['time'], DataStream.FromString(x['stream']).encode(), x['value'], reading_id=x['reading_id']) for x in in_data['trace']]

        return SimulationTrace(readings, selectors=selectors)



from .scope import Scope
from ...stream import DataStream
from ...slot import SlotIdentifier
from ...node import InputTrigger
from ...exceptions import SensorGraphSemanticError
from ...known_constants import (system_tick, fast_tick, config_fast_tick_secs, tick_1,
                                config_tick1_secs, tick_2, config_tick2_secs)


class RootScope(Scope):
    """The global scope that contains all others.

    Args:
        sensor_graph (SensorGraph): The SensorGraph we
            are operating on.
        alloc (StreamAllocator): The global stream allocator we
            are using.
    """

    def __init__(self, sensor_graph, alloc):
        super(RootScope, self).__init__(u"Root Scope", sensor_graph, alloc, None)

        self.system_tick = None
        self.fast_tick = None
        self.user1_tick = None
        self.user2_tick = None

        self._setup()

    def _setup(self):
        """Prepare for code generation by setting up root clock nodes.

        These nodes are subsequently used as the basis for all clock operations.
        """

        # Create a root system ticks and user configurable ticks
        systick = self.allocator.allocate_stream(DataStream.CounterType, attach=True)
        fasttick = self.allocator.allocate_stream(DataStream.CounterType, attach=True)
        user1tick = self.allocator.allocate_stream(DataStream.CounterType, attach=True)
        user2tick = self.allocator.allocate_stream(DataStream.CounterType, attach=True)

        self.sensor_graph.add_node("({} always) => {} using copy_all_a".format(system_tick, systick))
        self.sensor_graph.add_node("({} always) => {} using copy_all_a".format(fast_tick, fasttick))
        self.sensor_graph.add_config(SlotIdentifier.FromString('controller'), config_fast_tick_secs, 'uint32_t', 1)

        self.sensor_graph.add_node("({} always) => {} using copy_all_a".format(tick_1, user1tick))
        self.sensor_graph.add_node("({} always) => {} using copy_all_a".format(tick_2, user2tick))
        self.system_tick = systick
        self.fast_tick = fasttick
        self.user1_tick = user1tick
        self.user2_tick = user2tick

    def trigger_chain(self):
        """Return a NodeInput tuple for creating a node.

        Returns:
            (StreamIdentifier, InputTrigger)
        """

        raise SensorGraphSemanticError("There is no trigger chain in the root scope since no triggering criteria have been set")

    def clock(self, interval, basis="system"):
        """Return a NodeInput tuple for triggering an event every interval.

        Args:
            interval (int): The interval at which this input should
                trigger. If basis == system (the default), this interval must
                be in seconds.  Otherwise it will be in units of whatever the
                basis tick is configured with.
            basis (str): The basis to use for calculating the interval.  This
                can either be system, tick_1 or tick_2.  System means that the
                clock will use either the fast or regular builtin tick.  Passing
                tick_1 or tick_2 will cause the clock to be generated based on
                the selected tick.
        """

        if basis == "system":
            if (interval % 10) == 0:
                tick = self.allocator.attach_stream(self.system_tick)
                count = interval // 10
            else:
                tick = self.allocator.attach_stream(self.fast_tick)
                count = interval

            trigger = InputTrigger(u'count', '>=', count)
            return (tick, trigger)
        elif basis == 'tick_1':
            tick = self.allocator.attach_stream(self.user1_tick)
            trigger = InputTrigger(u'count', '>=', interval)
            return (tick, trigger)
        elif basis == 'tick_2':
            tick = self.allocator.attach_stream(self.user2_tick)
            trigger = InputTrigger(u'count', '>=', interval)
            return (tick, trigger)

        raise SensorGraphSemanticError("Unkwown tick source specified in RootScope.clock", basis=basis)



"""Configuration block for assigning config variables to a tile."""

from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from ..scopes.scope import Scope


@python_2_unicode_compatible
class ConfigBlock(SensorGraphStatement):
    """A block of config variables to assign to a tile.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        children(list(SensorGraphStatement)): The statements that are
            part of this config block.
        location (LocationInfo): A namedtuple with information on the line this
            statement was generated from so that we can log appropriate error
            messages.
    """

    def __init__(self, parsed, children, location=None):
        self.slot = parsed[0]

        super(ConfigBlock, self).__init__(children, location)

    def __str__(self):
        return u"config {}".format(self.slot)

    def execute_before(self, sensor_graph, scope_stack):
        """Execute statement before children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]
        new_scope = Scope("Configuration Scope", sensor_graph, parent.allocator, parent)
        new_scope.add_identifier('current_slot', self.slot)
        scope_stack.append(new_scope)

    def execute_after(self, sensor_graph, scope_stack):
        """Execute statement after children are executed.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        scope_stack.pop()



"""Convert copy_all to copy_latest whenever we can."""

from iotile.core.exceptions import ArgumentError
from iotile.sg.node import TrueTrigger, FalseTrigger, InputTrigger
from iotile.sg import DataStreamSelector, DataStream
import copy


class ConvertCopyAllToCopyLatest(object):
    """Run the copy_all to copy_latest optimization

    Copy_latest is much easier to optimize because we know it's always
    going to produce either 0 or 1 reading on the output.  Copy all on
    the other hand could produce many readings.

    There are some cases when we know that copy_all can be downgraded to
    copy_latest.

    The cases where we currently detect the operations as being the same
    are:

    - For unbuffered streams that are not counters,
      they are the same (i.e. input, unbuffered).
    - For counter streams whose selector is count == 1, they are the same
    - For counter streams whose selectror is always, they are the same
      since the counter stream starts with no readings and every time a
      reading is added

    Args:
        sensor_graph (SensorGraph): The sensor graph to run
            the optimization pass on
    """

    def __init__(self):
        pass

    def run(self, sensor_graph, model):
        """Run this optimization pass on the sensor graph

        If necessary, information on the device model being targeted
        can be found in the associated model argument.

        Args:
            sensor_graph (SensorGraph): The sensor graph to optimize
            model (DeviceModel): The device model we're using
        """

        # This check can be done if there is 1 input and it is count == 1
        # and the stream type is input or unbuffered

        did_downgrade = False

        for node, inputs, _outputs in sensor_graph.iterate_bfs():
            can_downgrade = False

            if node.func_name != u'copy_all_a':
                continue

            input_a, trigger_a = node.inputs[0]

            # We can always downgrade unbuffered non-counter
            if input_a.selector.match_type in (DataStream.InputType, DataStream.UnbufferedType):
                can_downgrade = True
            elif isinstance(trigger_a, InputTrigger) and trigger_a.comp_string == u'==' and trigger_a.use_count and trigger_a.reference == 1:
                can_downgrade = True
            elif isinstance(trigger_a, TrueTrigger) and not input_a.selector.buffered:
                # We can only downgrade an always trigger if we know the node before this will not
                # generate more than one reading at a time.  Since in that case, we get readings one at a
                # time and we don't accumulate them so there will be at most one reading to copy.
                # So, we need to check the node that produces input A to this node and see if we can
                # prove that it will produce at most one reading at a time.
                can_downgrade = True
                for in_node in inputs:
                    if input_a.matches(in_node.stream) and in_node.func_name == u'copy_all_a' and in_node.input_a.match_type not in (DataStream.InputType, DataStream.UnbufferedType):
                        can_downgrade = False
                        break

            if can_downgrade:
                did_downgrade = True
                node.set_func(u'copy_latest_a', sensor_graph.find_processing_function(u'copy_latest_a'))

        return did_downgrade



"""A machine readable ascii format for storing config information from a sensor graph.

The iotile-sensorgraph package works on .sgf files which contain both data flow
definitions run using an embedded sensor graph engine on an IOTile Device and
static configuration information that defines how the IOTile device is configured.
"""

from iotile.core.utilities.command_file import CommandFile
from binascii import hexlify


def format_config(sensor_graph):
    """Extract the config variables from this sensor graph in ASCII format.

    Args:
        sensor_graph (SensorGraph): the sensor graph that we want to format

    Returns:
        str: The ascii output lines concatenated as a single string
    """

    cmdfile = CommandFile("Config Variables", "1.0")

    for slot, conf_vars in sensor_graph.config_database.items():
        for conf_var, conf_def in conf_vars.items():
            conf_type, conf_val = conf_def

            if conf_type == 'binary':
                conf_val = 'hex:' + hexlify(conf_val)

            cmdfile.add("set_variable", slot, conf_var, conf_type, conf_val)

    return cmdfile.dump()



"""Copy stream statement."""

from builtins import str
from future.utils import python_2_unicode_compatible
from .statement import SensorGraphStatement
from ...stream import DataStream


@python_2_unicode_compatible
class TriggerStatement(SensorGraphStatement):
    """Trigger a streamer

    The form of the statement should be
    trigger streamer <index>

    That streamer is manually triggered.

    Args:
        parsed(ParseResults): The parsed tokens that make up this
            statement.
        location (LocationInfo): A namedtuple with information on the line this
            statement was generated from so that we can log appropriate error
            messages.
    """

    def __init__(self, parsed, location=None):
        self.index = parsed['index']

        super(TriggerStatement, self).__init__([], location)

    def __str__(self):
        return u'trigger streamer {};'.format(self.index)

    def execute(self, sensor_graph, scope_stack):
        """Execute this statement on the sensor_graph given the current scope tree.

        This adds a single node to the sensor graph with the trigger_streamer function
        as is processing function.

        Args:
            sensor_graph (SensorGraph): The sensor graph that we are building or
                modifying
            scope_stack (list(Scope)): A stack of nested scopes that may influence
                how this statement allocates clocks or other stream resources.
        """

        parent = scope_stack[-1]
        alloc = parent.allocator

        # The output is unused
        output = alloc.allocate_stream(DataStream.UnbufferedType, attach=True)

        trigger_stream, trigger_cond = parent.trigger_chain()
        streamer_const = alloc.allocate_stream(DataStream.ConstantType, attach=True)

        sensor_graph.add_node(u"({} {} && {} always) => {} using trigger_streamer".format(trigger_stream, trigger_cond, streamer_const, output))
        sensor_graph.add_constant(streamer_const, self.index)



"""A centralized data management structure that allows multiple named FIFO streams.

SensorGraph is organized around data streams, which are named FIFOs that hold
readings.  For the purposes of embedding SensorGraph into a low resource setting
you can limit the number of readings stored in any given FIFO as needed to put
a hard cap on storage requirements.
"""

import copy
from .engine import InMemoryStorageEngine
from .stream import DataStreamSelector, DataStream
from .walker import VirtualStreamWalker, CounterStreamWalker, BufferedStreamWalker
from .exceptions import StreamEmptyError, StorageFullError
from iotile.sg.model import DeviceModel
from iotile.core.exceptions import ArgumentError


class SensorLog(object):
    """A storage engine holding multiple named FIFOs.

    Args:
        engine (StorageEngine): The engine used for storing
            persistent data generated by this sensor graph.
            This can either be a simple, in memory data store, or
            a more complicated persistent storage setup depending on
            needs.  If not specified, a temporary in memory engine
            is used.

        model (DeviceModel): An optional device model specifying the
            constraints of the device that we are emulating.  If not
            specified, no specific constraints are imposed on the sensor
            log.
    """

    def __init__(self, engine=None, model=None):
        self._monitors = {}
        self._last_values = {}
        self._virtual_walkers = []
        self._queue_walkers = []

        if model is None:
            model = DeviceModel()

        if engine is None:
            engine = InMemoryStorageEngine(model=model)

        self._engine = engine
        self._model = model

    def watch(self, selector, callback):
        """Call a function whenever a stream changes.

        Args:
            selector (DataStreamSelector): The selector to watch
            callback (callable): The function to call when a new
                reading is pushed.  Callback is called as:
                callback(stream, value)
        """

        if selector not in self._monitors:
            self._monitors[selector] = set()

        self._monitors[selector].add(callback)

    def create_walker(self, selector):
        """Create a stream walker based on the given selector.
`
        This function returns a StreamWalker subclass that will
        remain up to date and allow iterating over and popping readings
        from the stream(s) specified by the selector.

        When the stream walker is done, it should be passed to
        destroy_walker so that it is removed from internal lists that
        are used to always keep it in sync.

        Args:
            selector (DataStreamSelector): The selector describing the
                streams that we want to iterate over.
        """

        if selector.buffered:
            walker = BufferedStreamWalker(selector, self._engine)
            self._queue_walkers.append(walker)
            return walker

        if selector.match_type == DataStream.CounterType:
            walker = CounterStreamWalker(selector)
        else:
            walker = VirtualStreamWalker(selector)

        self._virtual_walkers.append(walker)

        return walker

    def destroy_walker(self, walker):
        """Destroy a previously created stream walker.

        Args:
            walker (StreamWalker): The walker to remove from internal updating
                lists.
        """

        if walker.buffered:
            self._queue_walkers.remove(walker)
        else:
            self._virtual_walkers.remove(walker)

    def clear(self):
        """Clear all data from this sensor_log.

        All readings in all walkers are skipped and buffered data is
        destroyed.
        """

        for walker in self._virtual_walkers:
            walker.skip_all()

        self._engine.clear()

        for walker in self._queue_walkers:
            walker.skip_all()

    def push(self, stream, reading):
        """Push a reading into a stream, updating any associated stream walkers.

        Args:
            stream (DataStream): the stream to push the reading into
            reading (IOTileReading): the reading to push
        """

        # Make sure the stream is correct
        reading = copy.copy(reading)
        reading.stream = stream.encode()

        if stream.buffered:
            output_buffer = stream.output

            try:
                self._engine.push(reading)
            except StorageFullError:
                self._erase_buffer(stream.output)
                self._engine.push(reading)

            for walker in self._queue_walkers:
                # Only notify the walkers that are on this queue
                if walker.selector.output == output_buffer:
                    walker.notify_added(stream)

        # Activate any monitors we have for this stream
        for selector in self._monitors:
            if selector.matches(stream):
                for callback in self._monitors[selector]:
                    callback(stream, reading)

        # Virtual streams live only in their walkers, so update each walker
        # that contains this stream.
        for walker in self._virtual_walkers:
            if walker.matches(stream):
                walker.push(stream, reading)

        self._last_values[stream] = reading

    def _erase_buffer(self, output_buffer):
        """Erase readings in the specified buffer to make space."""

        erase_size = self._model.get(u'buffer_erase_size')

        buffer_type = u'storage'
        if output_buffer:
            buffer_type = 'streaming'

        old_readings = self._engine.popn(buffer_type, erase_size)

        # Now go through all of our walkers that could match and
        # update their availability counts and data buffer pointers
        for reading in old_readings:
            stream = DataStream.FromEncoded(reading.stream)

            for walker in self._queue_walkers:
                # Only notify the walkers that are on this queue
                if walker.selector.output == output_buffer:
                    walker.notify_rollover(stream)

    def inspect_last(self, stream):
        """Return the last value pushed into a stream.

        This function works even if the stream is virtual and no
        virtual walker has been created for it.  It is primarily
        useful to aid in debugging sensor graphs.

        Args:
            stream (DataStream): The stream to inspect

        Returns:
            IOTileReading: The data in the stream

        Raises:
            StreamEmptyError: if there has never been data written to
                the stream.
        """

        if stream in self._last_values:
            return self._last_values[stream]

        raise StreamEmptyError(u"inspect_last called on stream that has never been written to", stream=stream)



from pyparsing import Optional, Word, Literal, Forward, alphas, nums, Group, Regex, ZeroOrMore, oneOf, restOfLine, QuotedString, StringEnd, Empty
from binascii import unhexlify
from ..stream import DataStream, DataStreamSelector
from ..slot import SlotIdentifier


ident = None
rvalue = None
number = None
binary = None
comp = None
config_type = None
slot_id = None
stream = None
selector = None
stream_trigger = None
time_interval = None
tick_interval = None
quoted_string = None
semi = None
comment = None

streamer_stmt = None

callrpc_stmt = None
subtract_stmt = None
simple_statement = None
generic_statement = None

block_id = None
block_bnf = None
statement = None
sensor_graph = None


def _create_primitives():
    global binary, ident, rvalue, number, quoted_string, semi, tick_interval, time_interval, slot_id, comp, config_type, stream, comment, stream_trigger, selector

    if ident is not None:
        return

    semi = Literal(u';').suppress()
    ident = Word(alphas+u"_", alphas + nums + u"_")
    number = Regex(u'((0x[a-fA-F0-9]+)|[+-]?[0-9]+)').setParseAction(lambda s, l, t: [int(t[0], 0)])
    binary = Regex(u'hex:([a-fA-F0-9][a-fA-F0-9])+').setParseAction(lambda s, l, t: [unhexlify(t[0][4:])])
    quoted_string = QuotedString(quoteChar='"', escChar="\\", unquoteResults=True)

    comment = Literal('#') + restOfLine

    rvalue = number | quoted_string

    # Convert all time intervals into an integer number of seconds
    time_unit_multipliers = {
        u'second': 1,
        u'seconds': 1,
        u'minute': 60,
        u'minutes': 60,
        u'hour': 60*60,
        u'hours': 60*60,
        u'day': 60*60*24,
        u'days': 60*60*24,
        u'month': 60*60*24*30,
        u'months': 60*60*24*30,
        u'year': 60*60*24*365,
        u'years': 60*60*24*365,
    }

    config_type = oneOf('uint8_t uint16_t uint32_t int8_t int16_t int32_t uint8_t[] uint16_t[] uint32_t[] int8_t[] int16_t[] int32_t[] string binary')
    comp = oneOf('> < >= <= == ~=')

    # Time intervals are all based on internal system clocks so we include a 'system' tag
    time_unit = oneOf(u"second seconds minute minutes hour hours day days week weeks month months year years")
    time_interval = (number + time_unit).setParseAction(lambda s, l, t: [t[0]*time_unit_multipliers[t[1]], 'system'])
    tick_interval = (number + (Literal("tick_1") | Literal("tick_2"))).setParseAction(lambda s, l, t: [t[0], t[1]])

    slot_id = Literal(u"controller") | (Literal(u'slot') + number)
    slot_id.setParseAction(lambda s,l,t: [SlotIdentifier.FromString(u' '.join([str(x) for x in t]))])

    stream_modifier = Literal("system") | Literal("user") | Literal("combined")

    stream = Optional(Literal("system")) + oneOf("buffered unbuffered input output counter constant") + number + Optional(Literal("node"))
    stream.setParseAction(lambda s,l,t: [DataStream.FromString(u' '.join([str(x) for x in t]))])

    all_selector = Optional(Literal("all")) + Optional(stream_modifier) + oneOf("buffered unbuffered inputs outputs counters constants") + Optional(Literal("nodes"))
    all_selector.setParseAction(lambda s,l,t: [DataStreamSelector.FromString(u' '.join([str(x) for x in t]))])
    one_selector = Optional(Literal("system")) + oneOf("buffered unbuffered input output counter constant") + number + Optional(Literal("node"))
    one_selector.setParseAction(lambda s,l,t: [DataStreamSelector.FromString(u' '.join([str(x) for x in t]))])

    selector = one_selector | all_selector

    trigger_comp = oneOf('> < >= <= ==')
    stream_trigger = Group((Literal(u'count') | Literal(u'value')) + Literal(u'(').suppress() - stream - Literal(u')').suppress() - trigger_comp - number).setResultsName('stream_trigger')


def _create_simple_statements():
    global binary, ident, rvalue, simple_statement, semi, comp, number, slot_id, subtract_stmt, callrpc_stmt, generic_statement, streamer_stmt, stream, selector

    if simple_statement is not None:
        return

    meta_stmt = Group(Literal('meta').suppress() + ident + Literal('=').suppress() + rvalue + semi).setResultsName('meta_statement')
    require_stmt = Group(Literal('require').suppress() + ident + comp + rvalue + semi).setResultsName('require_statement')
    set_stmt = Group(Literal('set').suppress() - (ident | number) - Literal("to").suppress() - (rvalue | binary) - Optional(Literal('as').suppress() + config_type) + semi).setResultsName('set_statement')
    callrpc_stmt = Group(Literal("call").suppress() + (ident | number) + Literal("on").suppress() + slot_id + Optional(Literal("=>").suppress() + stream('explicit_stream')) + semi).setResultsName('call_statement')
    streamer_stmt = Group(Optional(Literal("manual")('manual')) + Optional(oneOf(u'encrypted signed')('security')) + Optional(Literal(u'realtime')('realtime')) + Literal('streamer').suppress() -
                          Literal('on').suppress() - selector('selector') - Optional(Literal('to').suppress() - slot_id('explicit_tile')) - Optional(Literal('with').suppress() - Literal('streamer').suppress() - number('with_other')) - semi).setResultsName('streamer_statement')
    copy_stmt = Group(Literal("copy").suppress() - Optional(oneOf("all count average")('modifier')) - Optional(stream('explicit_input') | number('constant_input')) - Literal("=>").suppress() - stream("output") - semi).setResultsName('copy_statement')
    subtract_stmt = Group(Literal("subtract").suppress() - stream('subtract_input') - Literal('=>').suppress() - stream('stream') - Optional(Literal(",").suppress() - Literal('default') - number('default'))).setResultsName('subtract_statement')
    trigger_stmt = Group(Literal("trigger") - Literal("streamer") - number('index') - semi).setResultsName('trigger_statement')

    simple_statement = meta_stmt | require_stmt | set_stmt | callrpc_stmt | streamer_stmt | trigger_stmt | copy_stmt | subtract_stmt

    # In generic statements, keep track of the location where the match started for error handling
    locator = Empty().setParseAction(lambda s, l, t: l)('location')
    generic_statement = Group(locator + Group(ZeroOrMore(Regex(u"[^{};]+")) + Literal(u';'))('match')).setResultsName('unparsed_statement')


def _create_block_bnf():
    global block_bnf, time_interval, slot_id, statement, block_id, ident, stream

    if block_bnf is not None:
        return

    trigger_clause = Group(stream_trigger | Group(stream).setResultsName('stream_always') | Group(ident).setResultsName('identifier'))

    every_block_id = Group(Literal(u'every').suppress() - (time_interval | tick_interval)).setResultsName('every_block')
    when_block_id = Group(Literal(u'when').suppress() + Literal("connected").suppress() - Literal("to").suppress() - slot_id).setResultsName('when_block')
    latch_block_id = Group(Literal(u'when').suppress() - stream_trigger).setResultsName('latch_block')
    config_block_id = Group(Literal(u'config').suppress() - slot_id).setResultsName('config_block')
    on_block_id = Group(Literal(u'on').suppress() - trigger_clause.setResultsName('triggerA') - Optional((Literal("and") | Literal("or")) - trigger_clause.setResultsName('triggerB'))).setResultsName('on_block')

    # Keep track of the location where the match started for error handling
    locator = Empty().setParseAction(lambda s, l, t: l)('location')
    block_id = Group(locator + (every_block_id | when_block_id | latch_block_id | config_block_id | on_block_id))

    block_bnf = Forward()
    statement = generic_statement | block_bnf

    block_bnf << Group(block_id + Group(Literal(u'{').suppress() + ZeroOrMore(statement) + Literal(u'}').suppress())).setResultsName('block')


def get_language():
    """Create or retrieve the parse tree for defining a sensor graph."""

    global sensor_graph, statement

    if sensor_graph is not None:
        return sensor_graph

    _create_primitives()
    _create_simple_statements()
    _create_block_bnf()

    sensor_graph = ZeroOrMore(statement) + StringEnd()
    sensor_graph.ignore(comment)

    return sensor_graph


def get_statement():
    """Create or retrieve the parse tree for a simple statement."""

    global simple_statement

    get_language()

    return simple_statement



"""Time based stimulus class that can be applied to a simulation."""

from __future__ import (unicode_literals, absolute_import, print_function)
from builtins import str
from pyparsing import Optional, Literal, ParseException, ParseSyntaxException
from typedargs.exceptions import ArgumentError
from .stop_conditions import time_interval, number
import iotile.sg.parser.language as language

class SimulationStimulus(object):
    """A simulation stimulus that injects a value at a given time.

    Args:
        time (int): The time in 1 second ticks when this stimulus occurs
        stream (DataStream): The data stream that will be affected.  This
            must be an input stream.
        value (int): The value that will be injected to the given stream.
    """

    def __init__(self, time, stream, value):
        self.time = time
        self.stream = stream
        self.value = value

        if self.stream.stream_type != self.stream.InputType:
            raise ArgumentError("Invalid stimulus applied to non-input stream", stream=self.stream)

    @classmethod
    def FromString(cls, desc):
        """Create a new stimulus from a description string.

        The string must have the format:

        [time: ][system ]input X = Y
        where X and Y are integers.  The time, if given must
        be a time_interval, which is an integer followed by a
        time unit such as second(s), minute(s), etc.

        Args:
            desc (str): A string description of the stimulus.

        Returns:
            SimulationStimulus: The parsed stimulus object.
        """
        if language.stream is None:
            language.get_language()

        parse_exp = Optional(time_interval('time') - Literal(':').suppress()) - language.stream('stream') - Literal('=').suppress() - number('value')

        try:
            data = parse_exp.parseString(desc)
            time = 0
            if 'time' in data:
                time = data['time'][0]

            return SimulationStimulus(time, data['stream'][0], data['value'])
        except (ParseException, ParseSyntaxException):
            raise ArgumentError("Could not parse stimulus descriptor", descriptor=desc)



"""A device model that constrains what a sensor graph can do.

Any device specific settings like how many inputs or outputs a
sensor graph can have or what sorts of streams are supported should
be specified in the device model and then is used during sensor graph
construction and processing to make sure the graph could be supported
by the desired device type.
"""

from builtins import str
from iotile.core.exceptions import ArgumentError


class DeviceModel(object):
    """A model of a device's sensor graph processor used to constrain behavior."""

    def __init__(self):
        self._properties = {}

        self._add_property(u'max_node_inputs', 2)
        self._add_property(u'max_node_outputs', 4)
        self._add_property(u'max_root_nodes', 8)
        self._add_property(u'max_streamers', 8)
        self._add_property(u'max_nodes', 32)
        self._add_property(u'max_storage_buffer', 16128)
        self._add_property(u'max_streaming_buffer', 48896)
        self._add_property(u'buffer_erase_size', 256)

    def _add_property(self, name, default_value):
        """Add a device property with a given default value.

        Args:
            name (str): The name of the property to add
            default_value (int, bool): The value of the property
        """

        name = str(name)
        self._properties[name] = default_value

    def set(self, name, value):
        """Set a device model property.

        Args:
            name (str): The name of the property to set
            value (int, bool): The value of the property to set
        """

        name = str(name)
        if name not in self._properties:
            raise ArgumentError("Unknown property in DeviceModel", name=name)

        self._properties[name] = value

    def get(self, name):
        """Get a device model property.

        Args:
            name (str): The name of the property to get
        """

        name = str(name)
        if name not in self._properties:
            raise ArgumentError("Unknown property in DeviceModel", name=name)

        return self._properties[name]



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International
# are copyright Arch Systems Inc.

# Caveats and possible issues
# Mac OS X
# - when using a virtualenv, readline is not properly installed into the virtualenv
#   and cannot be imported.  You need to install it using easy_install as described here
#   http://calvinx.com/tag/readline/


from setuptools import setup, find_packages
import version

setup(
    name="iotile-sensorgraph",
    packages=find_packages(exclude=("test",)),
    version=version.version,
    license="LGPLv3",
    description="IOTile SensorGraph Management and Simulation Package",
    install_requires=[
        "pyparsing>=2.2.0",
        "future>=0.16.0",
        "monotonic>=1.3.0",
        "toposort>=1.5",
        "iotile-core>=3.16.7"
    ],
    entry_points={'iotile.sg_processor': ['copy_all_a = iotile.sg.processors:copy_all_a',
                                          'copy_latest_a = iotile.sg.processors:copy_latest_a',
                                          'copy_count_a = iotile.sg.processors:copy_count_a',
                                          'call_rpc = iotile.sg.processors:call_rpc',
                                          'trigger_streamer = iotile.sg.processors:trigger_streamer',
                                          'subtract_a_from_b = iotile.sg.processors:subtract_a_from_b'],
                  'console_scripts': ['iotile-sgrun = iotile.sg.scripts.iotile_sgrun:main',
                                      'iotile-sgcompile = iotile.sg.scripts.iotile_sgcompile:main']},
    author="Arch",
    author_email="info@arch-iot.com",
    url="https://github.com/iotile/coretools/iotilesensorgraph",
    keywords=["iotile", "arch", "embedded", "hardware"],
    classifiers=[
        "Programming Language :: Python",
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: GNU Library or Lesser General Public License (LGPL)",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules"
        ],
    long_description="""\
IOTileSensorGraph
-----------------

A package that parses, optimizes and runs sensor graph scripts.

See https://www.arch-iot.com.
"""
)



"""Tests to make sure the StreamAllocator works as intended."""

from iotile.sg.parser.stream_allocator import StreamAllocator
from iotile.sg import SensorGraph, DeviceModel, SensorLog, DataStream


def test_stream_allocation():
    """Make sure we can allocate DataStreams."""

    model = DeviceModel()
    log = SensorLog(model=model)
    sg = SensorGraph(log, model=model)

    # TODO Finish this function
    alloc = StreamAllocator(sg, model=model)

    stream1 = alloc.allocate_stream(DataStream.ConstantType)
    assert len(sg.nodes) == 0

    stream2 = alloc.attach_stream(stream1)
    assert len(sg.nodes) == 0

    stream3 = alloc.attach_stream(stream1)
    assert len(sg.nodes) == 0

    stream4 = alloc.attach_stream(stream1)
    assert len(sg.nodes) == 0

    stream5 = alloc.attach_stream(stream1)
    assert len(sg.nodes) == 1

    assert stream1 == stream2
    assert stream2 == stream3
    assert stream4 == stream1
    assert stream5 != stream1



"""Tests to make sure the iotile-sgrun command line tool is not broken."""


import sys
import os.path
import pytest
from iotile.sg.scripts.iotile_sgrun import main


class SystemExitError(Exception):
    """Internal error to signal that sys.exit was called."""
    pass

@pytest.fixture
def exitcode(monkeypatch):
    """Patch sys.exit so we can test argparse containing code."""

    def _fake_exit(code):
        raise SystemExitError(code)

    monkeypatch.setattr(sys, 'exit', _fake_exit)


def test_basic_help(exitcode):
    """Make sure the program can print help text."""

    with pytest.raises(SystemExitError):
        main(['--help'])


def test_basic_simulation(exitcode):
    """Make sure we can run a simulation to completion."""

    infile = os.path.join(os.path.dirname(__file__), 'sensor_graphs', 'basic_block.sgf')

    retval = main(['-s', 'run_time 1 second', infile])
    assert retval == 0



import pytest
from iotile.core.exceptions import ArgumentError
from iotile.sg.model import DeviceModel


def test_default_values():
    """Make sure we can get properties with default values."""

    model = DeviceModel()

    assert model.get('max_nodes') == 32
    assert model.get(u'max_nodes') == 32

    model.set('max_nodes', 16)
    assert model.get('max_nodes') == 16
    assert model.get(u'max_nodes') == 16

    model.set(u'max_nodes', 17)
    assert model.get('max_nodes') == 17
    assert model.get(u'max_nodes') == 17

    with pytest.raises(ArgumentError):
        model.get('unknown_parameter')

    with pytest.raises(ArgumentError):
        model.set('unknown_parameter', 15)



"""Command line script to load and run a sensor graph."""

import sys
import argparse
from builtins import str
from iotile.core.exceptions import ArgumentError, IOTileException
from iotile.sg import DeviceModel, DataStreamSelector, SlotIdentifier
from iotile.sg.sim import SensorGraphSimulator
from iotile.sg.sim.hosted_executor import SemihostedRPCExecutor
from iotile.sg.parser import SensorGraphFileParser
from iotile.sg.known_constants import user_connected
from iotile.sg.optimizer import SensorGraphOptimizer

DESCRIPTION = \
u"""Load and run a sensor graph, either in a simulator or on a physical device.

This program takes in a sensor graph description file and simulates it.  The
simulation can run without any physical device present, e.g. entirely on your
computer.  However, if you want you can also simulate the sensor graph in a
semihosted fashion, which means that the graph is still being evaluated on
your computer, but any RPCs it sends will be forwarded on to an actual device
for execution.

semihosting mode:
To activate semihosting mode you need to specify a -p option and a -d option.
The -p option is the port to use to connect to a device, and has the same format
as the --port argument given to the "iotile hw" command. The -d option is the
device id (also knows at the UUID of the device) and is an integer which can be
specified either decimal or hex (with a 0x prefix). See the examples section for
more details.

injecting stimuli:
Sometimes you need to be able to provide specific inputs to a sensor graph at
certain times in order to explore different potential execution paths.  Consider
for example, a sensor graph that only recorded data received on input 1. You
need to provide input 1 to your graph in order for it to do anything.  Use the
-i option to define stimuli.  You can specify as many as you want by using
multiple -i options.  The format for each is:

[number time_unit: ][system ]input X = Y
This specifies that the value Y on input X will be presented to the sensor graph
at the specified time.  If the time is omitted, it is set to the value 0 and
given at the start of the simulation.  Each stimulus is only applied once.  An
example of a valid stimulus is "1 minute: input 5 = 1", which means present the
value 1 on stream 'input 5' after 1 simulated minute (60 ticks).

examples:
    iotile-sgrun -p bled112 -d 25 <sensor_graph file>
        This will simulate the given sensor graph file on device id 25 that is
        connected via bluetooth using a bled112 bluetooth adapter.  There is
        no stop condition so the simulation will run until the user cancels it
        with a ctrl-c combination.

    iotile-sgrun -i "input 1 = 5" <sensor_graph file> -s "run_time 1 minute"
        This will run the simulation for exactly 60 simulated seconds and begin
        the simulation by injecting the value 5 onto input 1 exactly once.
"""


def build_args():
    """Create command line argument parser."""

    parser = argparse.ArgumentParser(description=DESCRIPTION, formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument(u'sensor_graph', type=str, help=u"The sensor graph file to load and run.")
    parser.add_argument(u'--stop', u'-s', action=u"append", default=[], type=str, help=u"A stop condition for when the simulation should end.")
    parser.add_argument(u'--realtime', u'-r', action=u"store_true", help=u"Do not accelerate the simulation, pin the ticks to wall clock time")
    parser.add_argument(u'--watch', u'-w', action=u"append", default=[], help=u"A stream to watch and print whenever writes are made.")
    parser.add_argument(u'--trace', u'-t', help=u"Trace all writes to output streams to a file")
    parser.add_argument(u'--disable-optimizer', action="store_true", help=u"disable the sensor graph optimizer completely")
    parser.add_argument(u"--mock-rpc", u"-m", action=u"append", type=str, default=[], help=u"mock an rpc, format should be <slot id>:<rpc_id> = value.  For example -m \"slot 1:0x500a = 10\"")
    parser.add_argument(u"--port", u"-p", help=u"The port to use to connect to a device if we are semihosting")
    parser.add_argument(u"--semihost-device", u"-d", type=lambda x: int(x, 0), help=u"The device id of the device we should semihost this sensor graph on.")
    parser.add_argument(u"-c", u"--connected", action="store_true", help=u"Simulate with a user connected to the device (to enable realtime outputs)")
    parser.add_argument(u"-i", u"--stimulus", action=u"append", default=[], help="Push a value to an input stream at the specified time (or before starting).  The syntax is [time: ][system ]input X = Y where X and Y are integers")
    return parser


def process_mock_rpc(input_string):
    """Process a mock RPC argument.

    Args:
        input_string (str): The input string that should be in the format
            <slot id>:<rpc id> = value
    """

    spec, equals, value = input_string.partition(u'=')

    if len(equals) == 0:
        print("Could not parse mock RPC argument: {}".format(input_string))
        sys.exit(1)

    try:
        value = int(value.strip(), 0)
    except ValueError as exc:
        print("Could not parse mock RPC value: {}".format(str(exc)))
        sys.exit(1)

    slot, part, rpc_id = spec.partition(u":")
    if len(part) == 0:
        print("Could not parse mock RPC slot/rpc definition: {}".format(spec))
        sys.exit(1)

    try:
        slot = SlotIdentifier.FromString(slot)
    except ArgumentError as exc:
        print("Could not parse slot id in mock RPC definition: {}".format(exc.msg))
        sys.exit(1)

    try:
        rpc_id = int(rpc_id, 0)
    except ValueError as exc:
        print("Could not parse mock RPC number: {}".format(str(exc)))
        sys.exit(1)

    return slot, rpc_id, value

def watch_printer(watch, value):
    """Print a watched value.

    Args:
        watch (DataStream): The stream that was watched
        value (IOTileReading): The value to was seen
    """

    print("({: 8} s) {}: {}".format(value.raw_time, watch, value.value))


def main(argv=None):
    """Main entry point for iotile sensorgraph simulator.

    This is the iotile-sgrun command line program.  It takes
    an optional set of command line parameters to allow for
    testing.

    Args:
        argv (list of str): An optional set of command line
            parameters.  If not passed, these are taken from
            sys.argv.
    """

    if argv is None:
        argv = sys.argv[1:]

    try:
        executor = None
        parser = build_args()
        args = parser.parse_args(args=argv)

        model = DeviceModel()

        parser = SensorGraphFileParser()
        parser.parse_file(args.sensor_graph)
        parser.compile(model)

        if not args.disable_optimizer:
            opt = SensorGraphOptimizer()
            opt.optimize(parser.sensor_graph, model=model)

        graph = parser.sensor_graph
        sim = SensorGraphSimulator(graph)

        for stop in args.stop:
            sim.stop_condition(stop)

        for watch in args.watch:
            watch_sel = DataStreamSelector.FromString(watch)
            graph.sensor_log.watch(watch_sel, watch_printer)

        # If we are semihosting, create the appropriate executor connected to the device
        if args.semihost_device is not None:
            executor = SemihostedRPCExecutor(args.port, args.semihost_device)
            sim.rpc_executor = executor

        for mock in args.mock_rpc:
            slot, rpc_id, value = process_mock_rpc(mock)
            sim.rpc_executor.mock(slot, rpc_id, value)

        for stim in args.stimulus:
            sim.stimulus(stim)

        graph.load_constants()

        if args.trace is not None:
            sim.record_trace()

        try:
            if args.connected:
                sim.step(user_connected, 8)

            sim.run(accelerated=not args.realtime)
        except KeyboardInterrupt:
            pass

        if args.trace is not None:
            sim.trace.save(args.trace)
    finally:
        if executor is not None:
            executor.hw.close()

    return 0



"""A SensorGraph simulator that can drive the sensor graph either in realtime or as fast as possible."""

import time
from monotonic import monotonic
from ..known_constants import system_tick, fast_tick, tick_1, tick_2, battery_voltage
from .null_executor import NullRPCExecutor
from .stop_conditions import TimeBasedStopCondition
from .trace import SimulationTrace
from .stimulus import SimulationStimulus
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.reports import IOTileReading

class SensorGraphSimulator(object):
    """A simulator for sensor graphs.

    At its core the simulator just sends timed inputs to the sensor graph
    every tick and then sits back and lets the nodes process the data in
    whatever way they are configure.

    There are a few things that need to be configured for the simulator to
    function properly however:

    1. There needs to be a stop condition.  When should the simulation
       stop running?  These conditions can either be simulated times like
       one day or other conditions like a certain number of readings.

    2. There needs to be a way to send RPCs.  Most interesting things inside
       of a sensor graph happen because it sent an RPC and processed its response.
       For this to work those RPCs need to be routed somewhere.  By default, all
       RPCs are accepted and a value of 0 is returned, however, you can configure
       the simulator to execute those RPCs on an actual device if you want.

       This results in a semi-hosted sensor graph, where the sensor graph is running
       on your computer but the rpcs are executed on another IOTile Device, which
       can be very useful for exploration and debugging.

    Args:
        sensor_graph (SensorGraph): The sensor graph that we want to simulate.
    """

    def __init__(self, sensor_graph):
        self.voltage = 3.6
        self.stop_conditions = []
        self.stimuli = []
        self._known_conditions = []
        self.trace = None
        self.tick_count = 0
        self.sensor_graph = sensor_graph
        self._start_tick = 0  # the tick on which the current simulation started
        self.rpc_executor = NullRPCExecutor()

        # Register known stop conditions
        self._known_conditions.append(TimeBasedStopCondition)

    def record_trace(self, selectors=None):
        """Record a trace of readings produced by this simulator.

        This causes the property `self.trace` to be populated with a
        SimulationTrace object that contains all of the readings that
        are produced during the course of the simulation.  Only readings
        that respond to specific selectors are given.

        You can pass whatever selectors you want in the optional selectors
        argument.  If you pass None, then the default behavior to trace
        all of the output streams of the sensor graph, which are defined
        as the streams that are selected by a DataStreamer object in the
        sensor graph.  This is typically what is meant by sensor graph
        output.

        You can inspect the current trace by looking at the trace
        property.  It will initially be an empty list and will be updated
        with each call to step() or run() that results in new readings
        responsive to the selectors you pick (or the graph streamers if
        you did not explicitly pass a list of DataStreamSelector objects).

        Args:
            selectors (list of DataStreamSelector): The selectors to add watch
                statements on to produce this trace. This is optional.
                If it is not specified, a the streamers of the sensor
                graph are used.
        """

        if selectors is None:
            selectors = [x.walker.selector for x in self.sensor_graph.streamers]

        self.trace = SimulationTrace(selectors=selectors)

        for sel in selectors:
            self.sensor_graph.sensor_log.watch(sel, self._on_trace_callback)

    def _on_trace_callback(self, watch, value):
        self.trace.append(value)

    def step(self, input_stream, value):
        """Step the sensor graph through one since input.

        The internal tick count is not advanced so this function may
        be called as many times as desired to input specific conditions
        without simulation time passing.

        Args:
            input_stream (DataStream): The input stream to push the
                value into
            value (int): The reading value to push as an integer
        """

        reading = IOTileReading(input_stream.encode(), self.tick_count, value)
        self.sensor_graph.process_input(input_stream, reading, self.rpc_executor)

    def run(self, include_reset=True, accelerated=True):
        """Run this sensor graph until a stop condition is hit.

        Multiple calls to this function are useful only if
        there has been some change in the stop conditions that would
        cause the second call to not exit immediately.

        Args:
            include_reset (bool): Start the sensor graph run with
                a reset event to match what would happen when an
                actual device powers on.
            accelerated (bool): Whether to run this sensor graph as
                fast as possible or to delay tick events to simulate
                the actual passage of wall clock time.
        """

        self._start_tick = self.tick_count

        if self._check_stop_conditions(self.sensor_graph):
            return

        if include_reset:
            pass  # TODO: include a reset event here

        # Process all stimuli that occur at the start of the simulation
        i = None
        for i, stim in enumerate(self.stimuli):
            if stim.time != 0:
                break

            reading = IOTileReading(self.tick_count, stim.stream.encode(), stim.value)
            self.sensor_graph.process_input(stim.stream, reading, self.rpc_executor)

        if i is not None and i > 0:
            self.stimuli = self.stimuli[i:]

        while not self._check_stop_conditions(self.sensor_graph):
            # Process one more one second tick
            now = monotonic()
            next_tick = now + 1.0

            # To match what is done in actual hardware, we increment tick count so the first tick
            # is 1.
            self.tick_count += 1

            # Process all stimuli that occur at this tick of the simulation
            i = None
            for i, stim in enumerate(self.stimuli):
                if stim.time != self.tick_count:
                    break

                reading = IOTileReading(self.tick_count, stim.stream.encode(), stim.value)
                self.sensor_graph.process_input(stim.stream, reading, self.rpc_executor)

            if i is not None and i > 0:
                self.stimuli = self.stimuli[i:]

            self._check_additional_ticks(self.tick_count)

            if (self.tick_count % 10) == 0:
                reading = IOTileReading(self.tick_count, system_tick.encode(), self.tick_count)
                self.sensor_graph.process_input(system_tick, reading, self.rpc_executor)

                # Every 10 seconds the battery voltage is reported in 16.16 fixed point format in volts
                reading = IOTileReading(self.tick_count, battery_voltage.encode(), int(self.voltage * 65536))
                self.sensor_graph.process_input(battery_voltage, reading, self.rpc_executor)

            now = monotonic()

            # If we are trying to execute this sensor graph in realtime, wait for
            # the remaining slice of this tick.
            if (not accelerated) and (now < next_tick):
                time.sleep(next_tick - now)

    def _check_additional_ticks(self, tick_value):
        fast_interval = self.sensor_graph.get_tick('fast')
        tick_1_interval = self.sensor_graph.get_tick('user1')
        tick_2_interval = self.sensor_graph.get_tick('user2')

        if fast_interval != 0 and (tick_value % fast_interval) == 0:
            reading = IOTileReading(self.tick_count, fast_tick.encode(), self.tick_count)
            self.sensor_graph.process_input(fast_tick, reading, self.rpc_executor)

        if tick_1_interval != 0 and (tick_value % tick_1_interval) == 0:
            reading = IOTileReading(self.tick_count, tick_1.encode(), self.tick_count)
            self.sensor_graph.process_input(tick_1, reading, self.rpc_executor)

        if tick_2_interval != 0 and (tick_value % tick_2_interval) == 0:
            reading = IOTileReading(self.tick_count, tick_2.encode(), self.tick_count)
            self.sensor_graph.process_input(tick_2, reading, self.rpc_executor)


    def _check_stop_conditions(self, sensor_graph):
        """Check if any of our stop conditions are met.

        Args:
            sensor_graph (SensorGraph): The sensor graph we are currently simulating

        Returns:
            bool: True if we should stop the simulation
        """

        for stop in self.stop_conditions:
            if stop.should_stop(self.tick_count, self.tick_count - self._start_tick, sensor_graph):
                return True

        return False

    def stimulus(self, stimulus):
        """Add a simulation stimulus at a given time.

        A stimulus is a specific input given to the graph at a specific
        time to a specific input stream.  The format for specifying a
        stimulus is:
        [time: ][system ]input X = Y
        where X and Y are integers.

        This will cause the simulator to inject this input at the given time.
        If you specify a time of 0 seconds, it will happen before the simulation
        starts.  Similarly, if you specify a time of 1 second it will also happen
        before anything else since the simulations start with a tick value of 1.

        The stimulus is injected before any other things happen at each new tick.

        Args:
            stimulus (str or SimulationStimulus): A prebuilt stimulus object or
                a string description of the stimulus of the format:
                [time: ][system ]input X = Y
                where time is optional and defaults to 0 seconds if not specified.

        Examples:
            sim.stimulus('system input 10 = 15')
            sim.stimulus('1 minute: input 1 = 5')
        """

        if not isinstance(stimulus, SimulationStimulus):
            stimulus = SimulationStimulus.FromString(stimulus)

        self.stimuli.append(stimulus)
        self.stimuli.sort(key=lambda x:x.time)

    def stop_condition(self, condition):
        """Add a stop condition to this simulation.

        Stop conditions are specified as strings and parsed into
        the appropriate internal structures.

        Args:
            condition (str): a string description of the stop condition
        """

        # Try to parse this into a stop condition with each of our registered
        # condition types
        for cond_format in self._known_conditions:
            try:
                cond = cond_format.FromString(condition)
                self.stop_conditions.append(cond)
                return
            except ArgumentError:
                continue

        raise ArgumentError("Stop condition could not be processed by any known StopCondition type", condition=condition, suggestion="It may be mistyped or otherwise invalid.")



"""A grammer for specifying sensor graph nodes."""

from __future__ import (unicode_literals, absolute_import, print_function)
from builtins import str
import struct
from pyparsing import Word, Regex, nums, hexnums, Literal, Optional, Group, oneOf, QuotedString, ParseException
from typedargs.exceptions import ArgumentError
from .node import SGNode, InputTrigger, FalseTrigger, TrueTrigger
from .stream import DataStream, DataStreamSelector

number = Regex('((0x[a-fA-F0-9]+)|[0-9]+)')
combiner = (Literal('&&') | Literal('||'))
symbol = Regex('[a-zA-Z][a-zA-Z_]*')

stream_type = Optional(Literal('system')) + (Literal('input') | Literal('output') | Literal('buffered') | Literal("unbuffered") | Literal("constant") | Literal("counter")) + Optional(Literal("node").suppress())
stream = stream_type + number

trigger_type = (Literal('value') | Literal('count'))
trigger_op = oneOf('> < >= <= ==')

trigger = Literal('always') | (Literal('when').suppress() + trigger_type('type') + trigger_op('op') + number('reference'))

inputstring = stream('input_stream') + trigger

inputdesc2 = Literal('(').suppress() + inputstring('input_a') + combiner('combiner') + inputstring('input_b') + Literal(')').suppress()
inputdesc1 = Literal('(').suppress() + inputstring('input_a') + Literal(')').suppress()

inputdesc = inputdesc1('input1') | inputdesc2('input2')
graph_node = inputdesc + Literal('=>').suppress() + stream('node') + Literal('using').suppress() + symbol('processor')


def parse_node_descriptor(desc, model):
    """Parse a string node descriptor.

    The function creates an SGNode object without connecting its inputs and outputs
    and returns a 3-tuple:

    SGNode, [(input X, trigger X)], <processing function name>

    Args:
        desc (str): A description of the node to be created.
        model (str): A device model for the node to be created that sets any
            device specific limits on how the node is set up.
    """

    try:
        data = graph_node.parseString(desc)
    except ParseException as exc:
        raise  # TODO: Fix this to properly encapsulate the parse error

    stream_desc = u' '.join(data['node'])

    stream = DataStream.FromString(stream_desc)
    node = SGNode(stream, model)

    inputs = []

    if 'input_a' in data:
        input_a = data['input_a']
        stream_a = DataStreamSelector.FromString(u' '.join(input_a['input_stream']))

        trigger_a = None
        if 'type' in input_a:
            trigger_a = InputTrigger(input_a['type'], input_a['op'], int(input_a['reference'], 0))

        inputs.append((stream_a, trigger_a))

    if 'input_b' in data:
        input_a = data['input_b']
        stream_a = DataStreamSelector.FromString(u' '.join(input_a['input_stream']))

        trigger_a = None
        if 'type' in input_a:
            trigger_a = InputTrigger(input_a['type'], input_a['op'], int(input_a['reference'], 0))

        inputs.append((stream_a, trigger_a))

    if 'combiner' in data and str(data['combiner']) == u'&&':
        node.trigger_combiner = SGNode.AndTriggerCombiner
    else:
        node.trigger_combiner = SGNode.OrTriggerCombiner

    processing = data['processor']
    return node, inputs, processing


def parse_binary_descriptor(bindata):
    """Convert a binary node descriptor into a string descriptor.

    Binary node descriptor are 20-byte binary structures that encode all
    information needed to create a graph node.  They are used to communicate
    that information to an embedded device in an efficent format.  This
    function exists to turn such a compressed node description back into
    an understandable string.

    Args:
        bindata (bytes): The raw binary structure that contains the node
            description.

    Returns:
        str: The corresponding string description of the same sensor_graph node
    """

    func_names = {0: 'copy_latest_a', 1: 'average_a',
                  2: 'copy_all_a', 3: 'sum_a',
                  4: 'copy_count_a', 5: 'trigger_streamer',
                  6: 'call_rpc', 7: 'subtract_afromb'}

    if len(bindata) != 20:
        raise ArgumentError("Invalid binary node descriptor with incorrect size", size=len(bindata), expected=20, bindata=bindata)

    a_trig, b_trig, stream_id, a_id, b_id, proc, a_cond, b_cond, trig_combiner = struct.unpack("<LLHHHBBBB2x", bindata)

    node_stream = DataStream.FromEncoded(stream_id)

    if a_id == 0xFFFF:
        raise ArgumentError("Invalid binary node descriptor with invalid first import", input_selector=a_id)

    a_selector = DataStreamSelector.FromEncoded(a_id)
    a_trigger = _process_binary_trigger(a_trig, a_cond)

    b_selector = None
    b_trigger = None
    if b_id != 0xFFFF:
        b_selector = DataStreamSelector.FromEncoded(b_id)
        b_trigger = _process_binary_trigger(b_trig, b_cond)

    if trig_combiner == SGNode.AndTriggerCombiner:
        comb = '&&'
    elif trig_combiner == SGNode.OrTriggerCombiner:
        comb = '||'
    else:
        raise ArgumentError("Invalid trigger combiner in binary node descriptor", combiner=trig_combiner)

    if proc not in func_names:
        raise ArgumentError("Unknown processing function", function_id=proc, known_functions=func_names)

    func_name = func_names[proc]

    # Handle one input nodes
    if b_selector is None:
        return '({} {}) => {} using {}'.format(a_selector, a_trigger, node_stream, func_name)

    return '({} {} {} {} {}) => {} using {}'.format(a_selector, a_trigger, comb,
                                                     b_selector, b_trigger,
                                                     node_stream, func_name)


def _process_binary_trigger(trigger_value, condition):
    """Create an InputTrigger object."""

    ops = {
        0: ">",
        1: "<",
        2: ">=",
        3: "<=",
        4: "==",
        5: 'always'
    }

    sources = {
        0: 'value',
        1: 'count'
    }

    encoded_source = condition & 0b1
    encoded_op = condition >> 1

    op = ops.get(encoded_op, None)
    source = sources.get(encoded_source, None)

    if op is None:
        raise ArgumentError("Unknown operation in binary trigger", condition=condition, operation=encoded_op, known_ops=ops)
    if source is None:
        raise ArgumentError("Unknown value source in binary trigger", source=source, known_sources=sources)

    if op == 'always':
        return TrueTrigger()

    return InputTrigger(source, op, trigger_value)



"""Wellknown streams that have global significance for all IOTile devices."""

from .stream import DataStream

# System streams
# These are streams that are generated automatically by an IOTile Device
# according to internal signals and events.  Some indicate abnormal conditions
# and others are standard signals.

# Battery voltage is sampled and reported every 10 seconds
battery_voltage = DataStream.FromString('system input 0')

# The system tick is fired every 10 seconds
system_tick = DataStream.FromString('system input 2')

# A system generated fast tick can be configured to any interval
# it is used internally by the sensograph compiler and typically
# set at 1 second.
fast_tick = DataStream.FromString('system input 3')

# These streams receive the address of the tile that an external
# user connected to or disconnected from.
user_connected = DataStream.FromString('system input 1025')
user_disconnected = DataStream.FromString('system input 1026')

tick_1 = DataStream.FromString('system input 5')
tick_2 = DataStream.FromString('system input 6')

# Known config variable ids on the controller

## An optional config variable that sets the time interval in seconds
## for the user tick input to sensor graph
config_fast_tick_secs = 0x2000

## Configuration for the time interval of tick_1 in seconds, it defaults
## to 0 if not specified
config_tick1_secs = 0x2002

## Configuration for the time interval of tick_2 in seconds, it defaults
## to 0 if not specified
config_tick2_secs = 0x2003
